import os
import tarfile
import tempfile
from inspect import signature
from typing import (
    Any,
    Dict,
    Union,
)
from zipfile import ZipFile

from bioblend.galaxy import dataset_collections
from . import GalaxyTestBase


class TestGalaxyDatasetCollections(GalaxyTestBase.GalaxyTestBase):
    def test_create_list_in_history(self):
        history_id = self.gi.histories.create_history(name="TestDSListCreate")["id"]
        dataset1_id = self._test_dataset(history_id)
        dataset2_id = self._test_dataset(history_id)
        dataset3_id = self._test_dataset(history_id)
        collection_response = self.gi.histories.create_dataset_collection(
            history_id=history_id,
            collection_description=dataset_collections.CollectionDescription(
                name="MyDatasetList",
                elements=[
                    dataset_collections.HistoryDatasetElement(name="sample1", id=dataset1_id),
                    dataset_collections.HistoryDatasetElement(name="sample2", id=dataset2_id),
                    dataset_collections.HistoryDatasetElement(name="sample3", id=dataset3_id),
                ],
            ),
            copy_elements=False,
        )
        assert collection_response["name"] == "MyDatasetList"
        assert collection_response["collection_type"] == "list"
        elements = collection_response["elements"]
        assert len(elements) == 3
        assert elements[0]["element_index"] == 0
        assert elements[0]["object"]["id"] == dataset1_id
        assert elements[1]["object"]["id"] == dataset2_id
        assert elements[2]["object"]["id"] == dataset3_id
        assert elements[2]["element_identifier"] == "sample3"

    def test_create_list_of_paired_datasets_in_history(self):
        history_id = self.gi.histories.create_history(name="TestDSListCreate")["id"]
        dataset1_id = self._test_dataset(history_id)
        dataset2_id = self._test_dataset(history_id)
        dataset3_id = self._test_dataset(history_id)
        dataset4_id = self._test_dataset(history_id)
        collection_response = self.gi.histories.create_dataset_collection(
            history_id=history_id,
            collection_description=dataset_collections.CollectionDescription(
                name="MyListOfPairedDatasets",
                type="list:paired",
                elements=[
                    dataset_collections.CollectionElement(
                        name="sample1",
                        type="paired",
                        elements=[
                            dataset_collections.HistoryDatasetElement(name="forward", id=dataset1_id),
                            dataset_collections.HistoryDatasetElement(name="reverse", id=dataset2_id),
                        ],
                    ),
                    dataset_collections.CollectionElement(
                        name="sample2",
                        type="paired",
                        elements=[
                            dataset_collections.HistoryDatasetElement(name="forward", id=dataset3_id),
                            dataset_collections.HistoryDatasetElement(name="reverse", id=dataset4_id),
                        ],
                    ),
                ],
            ),
            copy_elements=False,
        )
        assert collection_response["name"] == "MyListOfPairedDatasets"
        assert collection_response["collection_type"] == "list:paired"
        elements = collection_response["elements"]
        assert len(elements) == 2
        assert elements[0]["element_index"] == 0
        created_pair1 = elements[0]["object"]
        assert created_pair1["collection_type"] == "paired"
        assert len(created_pair1["elements"]) == 2
        forward_element1 = created_pair1["elements"][0]
        assert forward_element1["element_identifier"] == "forward"
        assert forward_element1["element_index"] == 0
        forward_dataset1 = forward_element1["object"]
        assert forward_dataset1["id"] == dataset1_id

        assert elements[1]["element_index"] == 1
        created_pair2 = elements[1]["object"]
        assert created_pair2["collection_type"] == "paired"
        assert len(created_pair2["elements"]) == 2
        reverse_element2 = created_pair2["elements"][1]
        reverse_dataset2 = reverse_element2["object"]

        assert reverse_element2["element_identifier"] == "reverse"
        assert reverse_element2["element_index"] == 1
        assert reverse_dataset2["id"] == dataset4_id

    def test_collections_in_history_index(self):
        history_id = self.gi.histories.create_history(name="TestHistoryDSIndex")["id"]
        history_dataset_collection = self._create_pair_in_history(history_id)
        contents = self.gi.histories.show_history(history_id, contents=True)
        assert len(contents) == 3
        assert contents[2]["id"] == history_dataset_collection["id"]
        assert contents[2]["name"] == "MyTestPair"
        assert contents[2]["collection_type"] == "paired"

    def test_show_history_dataset_collection(self):
        history_id = self.gi.histories.create_history(name="TestHistoryDSIndexShow")["id"]
        history_dataset_collection = self._create_pair_in_history(history_id)
        show_response = self.gi.histories.show_dataset_collection(history_id, history_dataset_collection["id"])
        for key in ["collection_type", "elements", "name", "deleted", "visible"]:
            assert key in show_response
        assert not show_response["deleted"]
        assert show_response["visible"]

    def test_delete_history_dataset_collection(self):
        history_id = self.gi.histories.create_history(name="TestHistoryDSDelete")["id"]
        history_dataset_collection = self._create_pair_in_history(history_id)
        self.gi.histories.delete_dataset_collection(history_id, history_dataset_collection["id"])
        show_response = self.gi.histories.show_dataset_collection(history_id, history_dataset_collection["id"])
        assert show_response["deleted"]

    def test_update_history_dataset_collection(self):
        history_id = self.gi.histories.create_history(name="TestHistoryDSDelete")["id"]
        history_dataset_collection = self._create_pair_in_history(history_id)
        self.gi.histories.update_dataset_collection(history_id, history_dataset_collection["id"], visible=False)
        show_response = self.gi.histories.show_dataset_collection(history_id, history_dataset_collection["id"])
        assert not show_response["visible"]

    def test_show_dataset_collection(self):
        history_id = self.gi.histories.create_history(name="TestDatasetCollectionShow")["id"]
        dataset_collection1 = self._create_pair_in_history(history_id)
        dataset_collection2 = self.gi.dataset_collections.show_dataset_collection(dataset_collection1["id"])
        for key in (
            "collection_type",
            "deleted",
            "id",
            "hid",
            "history_content_type",
            "history_id",
            "name",
            "url",
            "visible",
        ):
            assert dataset_collection1[key] == dataset_collection2[key]
        for element1, element2 in zip(dataset_collection1["elements"], dataset_collection2["elements"]):
            assert element1["id"] == element2["id"]
            assert element1.keys() == element2.keys()
            for key in element1["object"].keys():
                assert key in element2["object"].keys()

    def test_download_dataset_collection(self):
        history_id = self.gi.histories.create_history(name="TestDatasetCollectionDownload")["id"]
        dataset_collection_id = self._create_pair_in_history(history_id)["id"]
        self.gi.dataset_collections.wait_for_dataset_collection(dataset_collection_id)

        tempdir = tempfile.mkdtemp(prefix="bioblend_test_dataset_collection_download_")
        archive_path = os.path.join(tempdir, "dataset_collection")
        archive_type = self.gi.dataset_collections.download_dataset_collection(
            dataset_collection_id, file_path=archive_path
        )["archive_type"]
        expected_contents = signature(self._test_dataset).parameters["contents"].default + "\n"
        extract_dir_path = os.path.join(tempdir, "extracted_files")
        os.mkdir(extract_dir_path)

        if archive_type == "zip":
            archive: Union[ZipFile, tarfile.TarFile] = ZipFile(archive_path)
        elif archive_type == "tgz":
            archive = tarfile.open(archive_path)

        archive.extractall(extract_dir_path)
        for fname in os.listdir(extract_dir_path):
            dataset_dir_path = os.path.join(extract_dir_path, fname)
            file_path = os.path.join(dataset_dir_path, os.listdir(dataset_dir_path)[0])
            with open(file_path) as f:
                assert expected_contents == f.read()
        archive.close()

    def test_wait_for_dataset_collection(self):
        history_id = self.gi.histories.create_history(name="TestDatasetCollectionWait")["id"]
        dataset_collection_id = self._create_pair_in_history(history_id)["id"]
        dataset_collection = self.gi.dataset_collections.wait_for_dataset_collection(dataset_collection_id)
        for element in dataset_collection["elements"]:
            assert element["object"]["state"] == "ok"

    def _create_pair_in_history(self, history_id: str) -> Dict[str, Any]:
        dataset1_id = self._test_dataset(history_id)
        dataset2_id = self._test_dataset(history_id)
        collection_response = self.gi.histories.create_dataset_collection(
            history_id=history_id,
            collection_description=dataset_collections.CollectionDescription(
                name="MyTestPair",
                type="paired",
                elements=[
                    dataset_collections.HistoryDatasetElement(name="forward", id=dataset1_id),
                    dataset_collections.HistoryDatasetElement(name="reverse", id=dataset2_id),
                ],
            ),
            copy_elements=False,
        )
        return collection_response



import json
import os
import sys

import requests

# This example, provided for comparison with w5_metagenomics.py,
# contains the code required to run the metagenomics workflow
# *without* BioBlend.

URL = os.getenv("GALAXY_URL", "https://orione.crs4.it").rstrip("/")
API_URL = f"{URL}/api"
API_KEY = os.getenv("GALAXY_API_KEY")
if not API_KEY:
    sys.exit("API_KEY not set, see the README.txt file")
headers = {"Content-Type": "application/json", "x-api-key": API_KEY}

# Select "W5 - Metagenomics" from published workflows

workflow_name = "W5 - Metagenomics"
r = requests.get(f"{API_URL}/workflows", params={"show_published": True}, headers=headers)
workflows = r.json()
filtered_workflows = [_ for _ in workflows if _["published"] and _["name"] == workflow_name]
assert len(filtered_workflows) == 1
w = filtered_workflows[0]

# Import the workflow to user space

data = {"workflow_id": w["id"]}
r = requests.post(f"{API_URL}/workflows/import", data=json.dumps(data), headers=headers)
iw = r.json()
r = requests.get(f"{API_URL}/workflows/{iw['id']}", headers=headers)
iw_details = r.json()

# Select the "Orione SupMat" library

library_name = "Orione SupMat"
r = requests.get(f"{API_URL}/libraries", headers=headers)
libraries = r.json()
filtered_libraries = [_ for _ in libraries if _["name"] == library_name]
assert len(filtered_libraries) == 1
library = filtered_libraries[0]

# Select the "/Metagenomics/MetagenomicsDataset.fq" dataset

ds_name = "/Metagenomics/MetagenomicsDataset.fq"
r = requests.get(f"{API_URL}/libraries/{library['id']}/contents", headers=headers)
contents = r.json()
filtered_contents = [_ for _ in contents if _["type"] == "file" and _["name"] == ds_name]
assert len(filtered_contents) == 1
ld = filtered_contents[0]

# Select the blastn step

filtered_wf_steps = [_ for _ in iw_details["steps"].values() if _["tool_id"] and "blastn" in _["tool_id"]]
assert len(filtered_wf_steps) == 1
ws = filtered_wf_steps[0]
tool_id = ws["tool_id"]

# Get (a copy of) the parameters dict for the selected step

ws_parameters = ws["tool_inputs"].copy()
for k, v in ws_parameters.items():
    ws_parameters[k] = json.loads(v)

# Run the workflow on a new history with the selected dataset
# as input, setting the BLAST db to "16SMicrobial-20131106"

history_name = f"{workflow_name} output"
ws_parameters["db_opts"]["database"] = "16SMicrobial-20131106"
data = {
    "workflow_id": iw["id"],
    "parameters": {tool_id: {"db_opts": ws_parameters["db_opts"]}},
}
assert len(iw_details["inputs"]) == 1
input_step_id = iw_details["inputs"].keys()[0]
data["ds_map"] = {input_step_id: {"src": "ld", "id": ld["id"]}}
data["history"] = history_name
r = requests.post(f"{API_URL}/workflows", data=json.dumps(data), headers=headers)
r_dict = r.json()

print(f"Running workflow: {iw['name']} [{iw['id']}]")
print(f"Output history: {history_name} [{r_dict['history']}]")



from . import GalaxyTestBase


class TestGalaxyConfig(GalaxyTestBase.GalaxyTestBase):
    def test_get_config(self):
        response = self.gi.config.get_config()
        assert isinstance(response, dict)
        assert "brand" in response.keys()

    def test_get_version(self):
        response = self.gi.config.get_version()
        assert isinstance(response, dict)
        assert "version_major" in response.keys()

    def test_whoami(self):
        response = self.gi.config.whoami()
        assert isinstance(response, dict)
        assert "username" in response.keys()

    def test_reload_toolbox(self):
        response = self.gi.config.reload_toolbox()
        assert response is None



"""
Contains possible interactions with the Galaxy Tool data tables
"""

from typing import (
    Any,
    Dict,
    List,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class ToolDataClient(Client):
    module = "tool_data"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_data_tables(self) -> List[Dict[str, Any]]:
        """
        Get the list of all data tables.

        :rtype: list
        :return: A list of dicts with details on individual data tables.
          For example::

            [{"model_class": "TabularToolDataTable", "name": "fasta_indexes"},
             {"model_class": "TabularToolDataTable", "name": "bwa_indexes"}]
        """
        return self._get()

    def show_data_table(self, data_table_id: str) -> Dict[str, Any]:
        """
        Get details of a given data table.

        :type data_table_id: str
        :param data_table_id: ID of the data table

        :rtype: dict
        :return: A description of the given data table and its content.
          For example::

            {'columns': ['value', 'dbkey', 'name', 'path'],
             'fields': [['test id',
                         'test',
                         'test name',
                         '/opt/galaxy-dist/tool-data/test/seq/test id.fa']],
             'model_class': 'TabularToolDataTable',
             'name': 'all_fasta'}

        """
        return self._get(id=data_table_id)

    def reload_data_table(self, data_table_id: str) -> Dict[str, Any]:
        """
        Reload a data table.

        :type data_table_id: str
        :param data_table_id: ID of the data table

        :rtype: dict
        :return: A description of the given data table and its content.
          For example::

            {'columns': ['value', 'dbkey', 'name', 'path'],
             'fields': [['test id',
                         'test',
                         'test name',
                         '/opt/galaxy-dist/tool-data/test/seq/test id.fa']],
             'model_class': 'TabularToolDataTable',
             'name': 'all_fasta'}
        """
        url = self._make_url(data_table_id) + "/reload"
        return self._get(url=url)

    def delete_data_table(self, data_table_id: str, values: str) -> Dict[str, Any]:
        """
        Delete an item from a data table.

        :type data_table_id: str
        :param data_table_id: ID of the data table

        :type values: str
        :param values: a "|" separated list of column contents, there must be a
          value for all the columns of the data table

        :rtype: dict
        :return: Remaining contents of the given data table
        """
        payload = {"values": values}
        return self._delete(payload=payload, id=data_table_id)



import shutil
import tempfile

import pytest

from bioblend import (
    ConnectionError,
    galaxy,
)
from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyDatasets(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        self.history_id = self.gi.histories.create_history(name="TestDataset")["id"]
        self.dataset_contents = "line 1\nline 2\rline 3\r\nline 4"
        self.dataset_id = self._test_dataset(self.history_id, contents=self.dataset_contents)
        self.gi.datasets.wait_for_dataset(self.dataset_id)

    def tearDown(self):
        self.gi.histories.delete_history(self.history_id, purge=True)

    def test_show_nonexistent_dataset(self):
        with pytest.raises(ConnectionError):
            self.gi.datasets.show_dataset("nonexistent_id")

    def test_show_dataset(self):
        dataset = self.gi.datasets.show_dataset(self.dataset_id)
        assert dataset["id"] == self.dataset_id
        assert not dataset["deleted"]
        self.gi.histories.delete_dataset(self.history_id, self.dataset_id)
        dataset = self.gi.datasets.show_dataset(self.dataset_id)
        assert dataset["id"] == self.dataset_id
        assert dataset["deleted"]

    def test_download_dataset(self):
        with pytest.raises((TypeError, ConnectionError)):
            self.gi.datasets.download_dataset(None)  # type: ignore[call-overload]
        expected_contents = ("\n".join(self.dataset_contents.splitlines()) + "\n").encode()
        # download_dataset() with file_path=None is already tested in TestGalaxyTools.test_paste_content()
        # self._wait_and_verify_dataset(self.dataset_id, expected_contents)
        tempdir = tempfile.mkdtemp(prefix="bioblend_test_")
        try:
            downloaded_dataset = self.gi.datasets.download_dataset(
                self.dataset_id, file_path=tempdir, maxwait=GalaxyTestBase.BIOBLEND_TEST_JOB_TIMEOUT * 2
            )
            assert downloaded_dataset.startswith(tempdir)
            with open(downloaded_dataset, "rb") as f:
                assert f.read() == expected_contents
        finally:
            shutil.rmtree(tempdir)
        with tempfile.NamedTemporaryFile(prefix="bioblend_test_") as f:
            download_filename = self.gi.datasets.download_dataset(
                self.dataset_id,
                file_path=f.name,
                use_default_filename=False,
                maxwait=GalaxyTestBase.BIOBLEND_TEST_JOB_TIMEOUT,
            )
            assert download_filename == f.name
            f.flush()
            assert f.read() == expected_contents

    def test_get_datasets(self):
        datasets = self.gi.datasets.get_datasets()
        dataset_ids = [dataset["id"] for dataset in datasets]
        assert self.dataset_id in dataset_ids

    def test_get_datasets_history(self):
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id)
        assert len(datasets) == 1

    def test_get_datasets_limit_offset(self):
        datasets = self.gi.datasets.get_datasets(limit=1)
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, offset=1)
        assert datasets == []

    def test_get_datasets_name(self):
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, name="Pasted Entry")
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, name="Wrong Name")
        assert datasets == []

    @test_util.skip_unless_galaxy("release_20.05")
    def test_get_datasets_time(self):
        dataset = self.gi.datasets.show_dataset(self.dataset_id)
        ct = dataset["create_time"]
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, create_time_min=ct)
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, create_time_max=ct)
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, create_time_min="2100-01-01T00:00:00")
        assert datasets == []
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, create_time_max="2000-01-01T00:00:00")
        assert datasets == []

        ut = dataset["update_time"]
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, update_time_min=ut)
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, update_time_max=ut)
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, update_time_min="2100-01-01T00:00:00")
        assert datasets == []
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, update_time_max="2000-01-01T00:00:00")
        assert datasets == []

    @test_util.skip_unless_galaxy("release_20.05")
    def test_get_datasets_extension(self):
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, extension="txt")
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, extension="bam")
        assert datasets == []

    @test_util.skip_unless_galaxy("release_22.01")
    def test_get_datasets_extension_list(self):
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, extension=["bam", "txt"])
        assert len(datasets) == 1

    @test_util.skip_unless_galaxy("release_20.05")
    def test_get_datasets_state(self):
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, state="ok")
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, state="queued")
        assert datasets == []
        with pytest.raises(ConnectionError):
            self.gi.datasets.get_datasets(history_id=self.history_id, state="nonexistent_state")
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, state=["ok", "queued"])
        assert len(datasets) == 1

    @test_util.skip_unless_galaxy("release_20.05")
    def test_get_datasets_visible(self):
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, visible=True)
        assert len(datasets) == 1
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, visible=False)
        assert len(datasets) == 0

    def test_get_datasets_ordering(self):
        self.dataset_id2 = self._test_dataset(self.history_id, contents=self.dataset_contents)
        self.gi.datasets.wait_for_dataset(self.dataset_id2)
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, order="create_time-dsc")
        assert datasets[0]["id"] == self.dataset_id2
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, order="create_time-asc")
        assert datasets[0]["id"] == self.dataset_id
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, order="hid-dsc")
        assert datasets[0]["id"] == self.dataset_id2
        datasets = self.gi.datasets.get_datasets(history_id=self.history_id, order="hid-asc")
        assert datasets[0]["id"] == self.dataset_id

    def test_get_datasets_deleted(self):
        deleted_datasets = self.gi.datasets.get_datasets(history_id=self.history_id, deleted=True)
        assert deleted_datasets == []
        self.gi.histories.delete_dataset(self.history_id, self.dataset_id)
        deleted_datasets = self.gi.datasets.get_datasets(history_id=self.history_id, deleted=True)
        assert len(deleted_datasets) == 1
        purged_datasets = self.gi.datasets.get_datasets(history_id=self.history_id, purged=True)
        assert purged_datasets == []
        self.gi.histories.delete_dataset(self.history_id, self.dataset_id, purge=True, wait=True)
        purged_datasets = self.gi.datasets.get_datasets(history_id=self.history_id, purged=True)
        assert len(purged_datasets) == 1

    def test_get_datasets_tool_id_and_tag(self):
        cat1_datasets = self.gi.datasets.get_datasets(history_id=self.history_id, tool_id="cat1")
        assert cat1_datasets == []
        upload1_datasets = self.gi.datasets.get_datasets(history_id=self.history_id, tool_id="upload1")
        assert len(upload1_datasets) == 1
        self.gi.histories.update_dataset(self.history_id, self.dataset_id, tags=["test"])
        tagged_datasets = self.gi.datasets.get_datasets(history_id=self.history_id, tag="test")
        assert len(tagged_datasets) == 1

    def test_wait_for_dataset(self):
        history_id = self.gi.histories.create_history(name="TestWaitForDataset")["id"]
        dataset_contents = "line 1\nline 2\rline 3\r\nline 4"
        dataset_id = self._test_dataset(history_id, contents=dataset_contents)

        dataset = self.gi.datasets.wait_for_dataset(dataset_id)
        assert dataset["state"] == "ok"

        self.gi.histories.delete_history(history_id, purge=True)

    def test_dataset_permissions(self):
        admin_user_id = self.gi.users.get_current_user()["id"]
        username = test_util.random_string()
        user_id = self.gi.users.create_local_user(username, f"{username}@example.org", test_util.random_string(20))[
            "id"
        ]
        user_api_key = self.gi.users.create_user_apikey(user_id)
        anonymous_gi = galaxy.GalaxyInstance(url=self.gi.base_url, key=None)
        user_gi = galaxy.GalaxyInstance(url=self.gi.base_url, key=user_api_key)
        sharing_role = self.gi.roles.create_role("sharing_role", "sharing_role", [user_id, admin_user_id])["id"]

        self.gi.datasets.publish_dataset(self.dataset_id, published=False)
        with pytest.raises(ConnectionError):
            anonymous_gi.datasets.show_dataset(self.dataset_id)
        self.gi.datasets.publish_dataset(self.dataset_id, published=True)
        # now dataset is public, i.e. accessible to anonymous users
        assert anonymous_gi.datasets.show_dataset(self.dataset_id)["id"] == self.dataset_id
        self.gi.datasets.publish_dataset(self.dataset_id, published=False)

        with pytest.raises(ConnectionError):
            user_gi.datasets.show_dataset(self.dataset_id)
        self.gi.datasets.update_permissions(self.dataset_id, access_ids=[sharing_role], manage_ids=[sharing_role])
        assert user_gi.datasets.show_dataset(self.dataset_id)["id"] == self.dataset_id
        # anonymous access now fails because sharing is only with the shared user role
        with pytest.raises(ConnectionError):
            anonymous_gi.datasets.show_dataset(self.dataset_id)



"""
Contains possible interaction dealing with Galaxy users.

Most of these methods must be executed by a registered Galaxy admin user.
"""

from typing import (
    Any,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
)

from bioblend import ConnectionError
from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class UserClient(Client):
    module = "users"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_users(
        self,
        deleted: bool = False,
        f_email: Optional[str] = None,
        f_name: Optional[str] = None,
        f_any: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Get a list of all registered users. If ``deleted`` is set to ``True``,
        get a list of deleted users.

        :type deleted: bool
        :param deleted: Whether to include deleted users

        :type f_email: str
        :param f_email: filter for user emails. The filter will be active for
            non-admin users only if the Galaxy instance has the
            ``expose_user_email`` option set to ``true`` in the
            ``config/galaxy.yml`` configuration file.

        :type f_name: str
        :param f_name: filter for user names. The filter will be active for
            non-admin users only if the Galaxy instance has the
            ``expose_user_name`` option set to ``true`` in the
            ``config/galaxy.yml`` configuration file.

        :type f_any: str
        :param f_any: filter for user email or name. Each filter will be active
            for non-admin users only if the Galaxy instance has the
            corresponding ``expose_user_*`` option set to ``true`` in the
            ``config/galaxy.yml`` configuration file.

        :rtype: list
        :return: a list of dicts with user details.
                 For example::

                   [{'email': 'a_user@example.org',
                     'id': 'dda47097d9189f15',
                     'url': '/api/users/dda47097d9189f15'}]

        """
        params: Dict[str, Any] = {}
        if f_email:
            params["f_email"] = f_email
        if f_name:
            params["f_name"] = f_name
        if f_any:
            params["f_any"] = f_any
        return self._get(deleted=deleted, params=params)

    def show_user(self, user_id: str, deleted: bool = False) -> Dict[str, Any]:
        """
        Display information about a user.

        :type user_id: str
        :param user_id: encoded user ID

        :type deleted: bool
        :param deleted: whether to return results for a deleted user

        :rtype: dict
        :return: a dictionary containing information about the user
        """
        return self._get(id=user_id, deleted=deleted)

    def create_remote_user(self, user_email: str) -> Dict[str, Any]:
        """
        Create a new Galaxy remote user.

        .. note::
          This method works only if the Galaxy instance has the
          ``allow_user_creation`` and ``use_remote_user`` options set to
          ``true`` in the ``config/galaxy.yml`` configuration file. Also
          note that setting ``use_remote_user`` will require an upstream
          authentication proxy server; however, if you do not have one, access
          to Galaxy via a browser will not be possible.

        :type user_email: str
        :param user_email: email of the user to be created

        :rtype: dict
        :return: a dictionary containing information about the created user
        """
        payload = {
            "remote_user_email": user_email,
        }
        return self._post(payload)

    def create_local_user(self, username: str, user_email: str, password: str) -> Dict[str, Any]:
        """
        Create a new Galaxy local user.

        .. note::
          This method works only if the Galaxy instance has the
          ``allow_user_creation`` option set to ``true`` and
          ``use_remote_user`` option set to ``false`` in the
          ``config/galaxy.yml`` configuration file.

        :type username: str
        :param username: username of the user to be created

        :type user_email: str
        :param user_email: email of the user to be created

        :type password: str
        :param password: password of the user to be created

        :rtype: dict
        :return: a dictionary containing information about the created user
        """
        payload = {
            "username": username,
            "email": user_email,
            "password": password,
        }
        return self._post(payload)

    def get_current_user(self) -> Dict[str, Any]:
        """
        Display information about the user associated with this Galaxy
        connection.

        :rtype: dict
        :return: a dictionary containing information about the current user
        """
        url = self._make_url() + "/current"
        return self._get(url=url)

    def create_user_apikey(self, user_id: str) -> str:
        """
        Create a new API key for a given user.

        :type user_id: str
        :param user_id: encoded user ID

        :rtype: str
        :return: the API key for the user
        """
        url = self._make_url(user_id) + "/api_key"
        payload = {
            "user_id": user_id,
        }
        return self._post(payload, url=url)

    def delete_user(self, user_id: str, purge: bool = False) -> Dict[str, Any]:
        """
        Delete a user.

        .. note::
          This method works only if the Galaxy instance has the
          ``allow_user_deletion`` option set to ``true`` in the
          ``config/galaxy.yml`` configuration file.

        :type user_id: str
        :param user_id: encoded user ID

        :type purge: bool
        :param purge: if ``True``, also purge (permanently delete) the history

        :rtype: dict
        :return: a dictionary containing information about the deleted user
        """
        params = {}
        if purge is True:
            params["purge"] = purge
        return self._delete(id=user_id, params=params)

    def get_user_apikey(self, user_id: str) -> str:
        """
        Get the current API key for a given user.

        :type user_id: str
        :param user_id: encoded user ID

        :rtype: str
        :return: the API key for the user, or 'Not available.' if it doesn't
          exist yet.
        """
        try:
            url = self._make_url(user_id) + "/api_key/detailed"
            return self._get(url=url)["key"]
        except ConnectionError as e:
            if e.status_code == 204:
                return "Not available."
            elif e.status_code != 404:
                raise
            # Galaxy 22.05 or earlier
            url = self._make_url(user_id) + "/api_key/inputs"
            return self._get(url=url)["inputs"][0]["value"]

    def get_or_create_user_apikey(self, user_id: str) -> str:
        """
        Get the current API key for a given user, creating one if it doesn't
        exist yet.

        :type user_id: str
        :param user_id: encoded user ID

        :rtype: str
        :return: the API key for the user

        .. note::
          This method works only on Galaxy 21.01 or later.
        """
        url = self._make_url(user_id) + "/api_key"
        return self._get(url=url)

    def update_user(self, user_id: str, user_data: Optional[Dict] = None, **kwargs: Any) -> Dict[str, Any]:
        """
        Update user information. You can either pass the attributes you want to
        change in the user_data dictionary, or provide them separately as
        keyword arguments.
        For attributes that cannot be expressed as keywords (e.g.
        extra_user_preferences use a `|` sign), pass them in user_data.

        :type user_id: str
        :param user_id: encoded user ID

        :type user_data: dict
        :param user_data: a dict containing the values to be updated, eg. { "username" : "newUsername", "email": "new@email" }

        :type username: str
        :param username: Replace user name with the given string

        :type email: str
        :param email: Replace user email with the given string

        :rtype: dict
        :return: details of the updated user
        """
        if user_data is None:
            user_data = {}
        user_data.update(kwargs)
        url = self._make_url(user_id) + "/information/inputs"
        return self._put(url=url, payload=user_data, id=user_id)



import json
import os
import sys
from typing import (
    Any,
    Dict,
)

from common import get_one  # noqa:I100,I201

from bioblend.galaxy.objects import GalaxyInstance

URL = "https://orione.crs4.it"
API_KEY = os.getenv("GALAXY_API_KEY", "YOUR_API_KEY")
if API_KEY == "YOUR_API_KEY":
    sys.exit("API_KEY not set, see the README.txt file")
gi = GalaxyInstance(URL, api_key=API_KEY)

# Select "W3 - Bacterial de novo assembly | Paired-end" from published workflows

workflow_name = "W3 - Bacterial de novo assembly | Paired-end"
previews = gi.workflows.get_previews(name=workflow_name, published=True)
p = get_one(_ for _ in previews if _.published)

# Import the workflow to user space

iw = gi.workflows.import_shared(p.id)

# Create a new history

history_name = f"{workflow_name} output"
h = gi.histories.create(history_name)

# Select the "Orione SupMat" library

library_name = "Orione SupMat"
library = get_one(gi.libraries.list(name=library_name))

# Select the datasets

ds_names = [
    "/Whole genome - Escherichia coli/E coli DH10B MiSeq R1.fastq",
    "/Whole genome - Escherichia coli/E coli DH10B MiSeq R2.fastq",
]
input_labels = [
    "Left/Forward FASTQ Reads",
    "Right/Reverse FASTQ Reads",
]
input_map = {
    label: h.import_dataset(get_one(library.get_datasets(name=name))) for name, label in zip(ds_names, input_labels)
}

# Set the "hash_length" parameter to different values for the 3 "velveth" steps

lengths = {"19", "23", "29"}
ws_ids = iw.tool_labels_to_ids["velveth"]
assert len(ws_ids) == len(lengths)
params: Dict[str, Any] = {id_: {"hash_length": v} for id_, v in zip(ws_ids, lengths)}

# Set the "ins_length" runtime parameter to the same value for the 3
# "velvetg" steps

tool_id = "velvetg"
ws_ids = iw.tool_labels_to_ids[tool_id]
step = iw.steps[next(iter(ws_ids))]  # arbitrarily pick one
params[tool_id] = {"reads": json.loads(step.tool_inputs["reads"]).copy()}
params[tool_id]["reads"]["ins_length"] = -1

# Set more custom parameters

params["cisarunner"] = {"genomesize": 5000000}
params["check_contigs"] = {"genomesize": 5.0}
params["toolshed.g2.bx.psu.edu/repos/edward-kirton/abyss_toolsuite/abyss/1.0.0"] = {"k": 41}

# Run the workflow on a new history with the selected datasets as inputs

inv = iw.invoke(input_map, params=params, history=h, inputs_by="name")
out_hist = gi.histories.get(inv.history_id)
assert out_hist.name == history_name

print(f"Running workflow: {iw.name} [{iw.id}]")
print(f"Output history: {out_hist.name} [{out_hist.id}]")



"""
Interaction with a Tool Shed instance tools
"""

from typing import (
    Any,
    Dict,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.toolshed import ToolShedInstance


class ToolShedToolClient(Client):
    gi: "ToolShedInstance"
    module = "tools"

    def __init__(self, toolshed_instance: "ToolShedInstance") -> None:
        super().__init__(toolshed_instance)

    def search_tools(self, q: str, page: int = 1, page_size: int = 10) -> Dict[str, Any]:
        """
        Search for tools in a Galaxy Tool Shed.

        :type q: str
        :param q: query string for searching purposes

        :type page: int
        :param page: page requested

        :type page_size: int
        :param page_size: page size requested

        :rtype: dict
        :return: dictionary containing search hits as well as metadata for the
          search. For example::

            {'hits': [{'matched_terms': [],
                       'score': 3.0,
                       'tool': {'description': 'convert between various FASTQ quality formats',
                                'id': '69819b84d55f521efda001e0926e7233',
                                'name': 'FASTQ Groomer',
                                'repo_name': None,
                                'repo_owner_username': 'devteam'}},
                      {'matched_terms': [],
                       'score': 3.0,
                       'tool': {'description': 'converts a bam file to fastq files.',
                                'id': '521e282770fd94537daff87adad2551b',
                                'name': 'Defuse BamFastq',
                                'repo_name': None,
                                'repo_owner_username': 'jjohnson'}}],
             'hostname': 'https://testtoolshed.g2.bx.psu.edu/',
             'page': '1',
             'page_size': '2',
             'total_results': '118'}
        """
        params = {"q": q, "page": page, "page_size": page_size}
        return self._get(params=params)



""" General support infrastructure not tied to any particular test.
"""

import os
import random
import string
import unittest
from typing import (
    Callable,
    Optional,
)

import requests

import bioblend.galaxy

NO_GALAXY_MESSAGE = "Externally configured Galaxy required, but not found. Set BIOBLEND_GALAXY_URL and BIOBLEND_GALAXY_API_KEY to run this test."


def random_string(length: int = 8) -> str:
    return "".join(random.choice(string.ascii_lowercase) for _ in range(length))


def is_site_up(url: str) -> bool:
    try:
        response = requests.get(url, timeout=10)
        return response.status_code == 200
    except Exception:
        return False


def skip_unless_toolshed() -> Callable:
    """Decorate tests with this to skip the test if a URL for a ToolShed
    to run the tests is not provided.
    """
    if "BIOBLEND_TOOLSHED_URL" not in os.environ:
        return unittest.skip(
            "Externally configured ToolShed required, but not found. Set BIOBLEND_TOOLSHED_URL (e.g. to https://testtoolshed.g2.bx.psu.edu/ ) to run this test."
        )
    toolshed_url = os.environ["BIOBLEND_TOOLSHED_URL"]
    if not is_site_up(toolshed_url):
        return unittest.skip(f"Configured ToolShed [{toolshed_url}] appears to be down")
    return lambda f: f


def skip_unless_galaxy(min_release: Optional[str] = None) -> Callable:
    """Decorate tests with this to skip the test if Galaxy is not
    configured.
    """
    if min_release is not None:
        galaxy_release = os.environ.get("GALAXY_VERSION", None)
        if galaxy_release is not None and galaxy_release != "dev":
            if not galaxy_release.startswith("release_"):
                raise ValueError("The value of GALAXY_VERSION environment variable should start with 'release_'")
            if not min_release.startswith("release_"):
                raise Exception("min_release should start with 'release_'")
            if galaxy_release[8:] < min_release[8:]:
                return unittest.skip(f"Testing on Galaxy {galaxy_release}, but need {min_release} to run this test.")

    if "BIOBLEND_GALAXY_URL" not in os.environ:
        return unittest.skip(NO_GALAXY_MESSAGE)

    if "BIOBLEND_GALAXY_API_KEY" not in os.environ and "BIOBLEND_GALAXY_MASTER_API_KEY" in os.environ:
        galaxy_url = os.environ["BIOBLEND_GALAXY_URL"]
        galaxy_master_api_key = os.environ["BIOBLEND_GALAXY_MASTER_API_KEY"]
        gi = bioblend.galaxy.GalaxyInstance(galaxy_url, key=galaxy_master_api_key)

        if "BIOBLEND_GALAXY_USER_EMAIL" in os.environ:
            galaxy_user_email = os.environ["BIOBLEND_GALAXY_USER_EMAIL"]
        else:
            galaxy_user_email = f"{random_string()}@localhost.localdomain"

        galaxy_user_id = None
        for user in gi.users.get_users():
            if user["email"] == galaxy_user_email:
                galaxy_user_id = user["id"]
                break

        config = gi.config.get_config()
        if galaxy_user_id is None:
            if config.get("use_remote_user", False):
                new_user = gi.users.create_remote_user(galaxy_user_email)
            else:
                galaxy_user = galaxy_user_email.split("@", 1)[0]
                galaxy_password = random_string(20)

                # Create a new user
                new_user = gi.users.create_local_user(galaxy_user, galaxy_user_email, galaxy_password)
            galaxy_user_id = new_user["id"]

        if config["version_major"] >= "21.01":
            api_key = gi.users.get_or_create_user_apikey(galaxy_user_id)
        else:
            api_key = gi.users.get_user_apikey(galaxy_user_id)
            if not api_key or api_key == "Not available.":
                api_key = gi.users.create_user_apikey(galaxy_user_id)
        os.environ["BIOBLEND_GALAXY_API_KEY"] = api_key

    if "BIOBLEND_GALAXY_API_KEY" not in os.environ:
        return unittest.skip(NO_GALAXY_MESSAGE)

    return lambda f: f


def skip_unless_tool(tool_id: str) -> Callable:
    """Decorate a Galaxy test method as requiring a specific tool,
    skip the test case if the tool is unavailable.
    """

    def method_wrapper(method):
        def wrapped_method(has_gi, *args, **kwargs):
            tools = has_gi.gi.tools.get_tools()
            # In panels by default, so flatten out sections...
            tool_ids = [_["id"] for _ in tools]
            if tool_id not in tool_ids:
                raise unittest.SkipTest(f"Externally configured Galaxy instance requires tool {tool_id} to run test.")

            return method(has_gi, *args, **kwargs)

        # Must preserve method name so pytest can detect and report tests by
        # name.
        wrapped_method.__name__ = method.__name__
        return wrapped_method

    return method_wrapper


def get_abspath(path: str) -> str:
    return os.path.abspath(os.path.join(os.path.dirname(__file__), path))



"""
"""

import os
from typing import (
    Any,
    Dict,
)

from bioblend.galaxy.tools.inputs import (
    conditional,
    dataset,
    inputs,
    repeat,
)
from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyTools(GalaxyTestBase.GalaxyTestBase):
    def test_get_tools(self):
        # Test requires target Galaxy is configured with at least one tool.
        tools = self.gi.tools.get_tools()
        assert len(tools) > 0
        assert all(map(self._assert_is_tool_rep, tools))

    def test_get_tool_panel(self):
        # Test requires target Galaxy is configured with at least one tool
        # section.
        tool_panel = self.gi.tools.get_tool_panel()
        sections = [s for s in tool_panel if "elems" in s]
        assert len(sections) > 0
        assert all(map(self._assert_is_tool_rep, sections[0]["elems"]))

    def _assert_is_tool_rep(self, data):
        assert data["model_class"].endswith("Tool")
        # Special tools like SetMetadataTool may have different model_class
        # than Tool - but they all seem to end in tool.

        for key in ["name", "id", "version"]:
            assert key in data
        return True

    def test_paste_content(self):
        history = self.gi.histories.create_history(name="test_paste_data history")
        paste_text = "line 1\nline 2\rline 3\r\nline 4"
        tool_output = self.gi.tools.paste_content(paste_text, history["id"])
        assert len(tool_output["outputs"]) == 1
        # All lines in the resulting dataset should end with "\n"
        expected_contents = ("\n".join(paste_text.splitlines()) + "\n").encode()
        self._wait_and_verify_dataset(tool_output["outputs"][0]["id"], expected_contents)
        # Same with space_to_tab=True
        tool_output = self.gi.tools.paste_content(paste_text, history["id"], space_to_tab=True)
        assert len(tool_output["outputs"]) == 1
        expected_contents = ("\n".join("\t".join(_.split()) for _ in paste_text.splitlines()) + "\n").encode()
        self._wait_and_verify_dataset(tool_output["outputs"][0]["id"], expected_contents)

    def test_upload_file(self):
        history = self.gi.histories.create_history(name="test_upload_file history")

        fn = test_util.get_abspath("test_util.py")
        file_name = "test1"
        tool_output = self.gi.tools.upload_file(
            fn,
            # First param could be a regular path also of course...
            history_id=history["id"],
            file_name=file_name,
            dbkey="?",
            file_type="txt",
        )
        self._wait_for_and_verify_upload(tool_output, file_name, fn, expected_dbkey="?")

    def test_upload_file_dbkey(self):
        history = self.gi.histories.create_history(name="test_upload_file history")
        fn = test_util.get_abspath("test_util.py")
        file_name = "test1"
        dbkey = "hg19"
        tool_output = self.gi.tools.upload_file(
            fn,
            history_id=history["id"],
            file_name=file_name,
            dbkey=dbkey,
            file_type="txt",
        )
        self._wait_for_and_verify_upload(tool_output, file_name, fn, expected_dbkey=dbkey)

    @test_util.skip_unless_tool("random_lines1")
    def test_run_random_lines(self):
        # Run second test case from randomlines.xml
        history_id = self.gi.histories.create_history(name="test_run_random_lines history")["id"]
        with open(test_util.get_abspath(os.path.join("data", "1.bed"))) as f:
            contents = f.read()
        dataset_id = self._test_dataset(history_id, contents=contents)
        tool_inputs = (
            inputs()
            .set("num_lines", "1")
            .set("input", dataset(dataset_id))
            .set("seed_source", conditional().set("seed_source_selector", "set_seed").set("seed", "asdf"))
        )
        tool_output = self.gi.tools.run_tool(history_id=history_id, tool_id="random_lines1", tool_inputs=tool_inputs)
        assert len(tool_output["outputs"]) == 1
        # TODO: Wait for results and verify has 1 line and is
        # chr5  131424298   131424460   CCDS4149.1_cds_0_0_chr5_131424299_f 0   +

    @test_util.skip_unless_tool("cat1")
    def test_run_cat1(self):
        history_id = self.gi.histories.create_history(name="test_run_cat1 history")["id"]
        dataset1_id = self._test_dataset(history_id, contents="1 2 3")
        dataset2_id = self._test_dataset(history_id, contents="4 5 6")
        dataset3_id = self._test_dataset(history_id, contents="7 8 9")
        tool_inputs = (
            inputs()
            .set("input1", dataset(dataset1_id))
            .set(
                "queries",
                repeat()
                .instance(inputs().set("input2", dataset(dataset2_id)))
                .instance(inputs().set("input2", dataset(dataset3_id))),
            )
        )
        tool_output = self.gi.tools.run_tool(history_id=history_id, tool_id="cat1", tool_inputs=tool_inputs)
        assert len(tool_output["outputs"]) == 1
        # TODO: Wait for results and verify it has 3 lines - 1 2 3, 4 5 6,
        # and 7 8 9.

    @test_util.skip_unless_tool("CONVERTER_fasta_to_bowtie_color_index")
    def test_tool_dependency_install(self):
        installed_dependencies = self.gi.tools.install_dependencies("CONVERTER_fasta_to_bowtie_color_index")
        assert any(
            True for d in installed_dependencies if d.get("name") == "bowtie" and d.get("dependency_type") == "conda"
        ), f"installed_dependencies is {installed_dependencies}"
        status = self.gi.tools.uninstall_dependencies("CONVERTER_fasta_to_bowtie_color_index")
        assert status[0]["model_class"] == "NullDependency", status

    @test_util.skip_unless_tool("CONVERTER_fasta_to_bowtie_color_index")
    def test_tool_requirements(self):
        tool_requirements = self.gi.tools.requirements("CONVERTER_fasta_to_bowtie_color_index")
        assert any(
            True
            for tr in tool_requirements
            if {"dependency_type", "version"} <= set(tr.keys()) and tr.get("name") == "bowtie"
        ), f"tool_requirements is {tool_requirements}"

    @test_util.skip_unless_tool("CONVERTER_fasta_to_bowtie_color_index")
    def test_reload(self):
        response = self.gi.tools.reload("CONVERTER_fasta_to_bowtie_color_index")
        assert isinstance(response, dict)
        assert "message" in response
        assert "id" in response["message"]

    @test_util.skip_unless_tool("sra_source")
    def test_get_citations(self):
        citations = self.gi.tools.get_citations("sra_source")
        assert len(citations) == 2

    def _wait_for_and_verify_upload(
        self, tool_output: Dict[str, Any], file_name: str, fn: str, expected_dbkey: str = "?"
    ) -> None:
        assert len(tool_output["outputs"]) == 1
        output = tool_output["outputs"][0]
        assert output["name"] == file_name
        expected_contents = open(fn, "rb").read()
        self._wait_and_verify_dataset(output["id"], expected_contents)
        assert output["genome_build"] == expected_dbkey

    @test_util.skip_unless_tool("random_lines1")
    def test_get_tool_model(self):
        history_id = self.gi.histories.create_history(name="test_run_random_lines history")["id"]
        tool_model = self.gi.tools.build(tool_id="random_lines1", history_id=history_id)
        assert len(tool_model["inputs"]) == 3



"""
Test functions in bioblend.galaxy.tool_dependencies
"""

from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyToolDependencies(GalaxyTestBase.GalaxyTestBase):
    @test_util.skip_unless_galaxy("release_20.01")
    def test_summarize_toolbox(self):
        toolbox_summary = self.gi.tool_dependencies.summarize_toolbox()
        assert isinstance(toolbox_summary, list)
        assert len(toolbox_summary) > 0

        toolbox_summary_by_tool = self.gi.tool_dependencies.summarize_toolbox(index_by="tools")
        assert isinstance(toolbox_summary_by_tool, list)
        assert len(toolbox_summary_by_tool) > 0
        assert isinstance(toolbox_summary_by_tool[0], dict)
        assert "tool_ids" in toolbox_summary_by_tool[0]
        assert isinstance(toolbox_summary_by_tool[0]["tool_ids"], list)
        tool_id = toolbox_summary_by_tool[0]["tool_ids"][0]

        toolbox_summary_select_tool_ids = self.gi.tool_dependencies.summarize_toolbox(
            index_by="tools", tool_ids=[tool_id]
        )
        assert isinstance(toolbox_summary_select_tool_ids, list)
        assert len(toolbox_summary_select_tool_ids) == 1
        assert toolbox_summary_select_tool_ids[0]["tool_ids"][0] == tool_id

    @test_util.skip_unless_galaxy("release_20.01")
    def test_unused_dependency_paths(self):
        unused_paths = self.gi.tool_dependencies.unused_dependency_paths()
        assert isinstance(unused_paths, list)

    @test_util.skip_unless_galaxy("release_20.01")
    def test_delete_unused_dependency_paths(self):
        self.gi.tool_dependencies.delete_unused_dependency_paths(paths=[])



"""
Interaction with a Galaxy Tool Shed.
"""

from typing import (
    Any,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class ToolShedClient(Client):
    module = "tool_shed_repositories"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_repositories(self) -> List[Dict[str, Any]]:
        """
        Get the list of all installed Tool Shed repositories on this Galaxy instance.

        :rtype: list
        :return: a list of dictionaries containing information about
          repositories present in the Tool Shed.
          For example::

            [{'changeset_revision': '4afe13ac23b6',
              'deleted': False,
              'dist_to_shed': False,
              'error_message': '',
              'name': 'velvet_toolsuite',
              'owner': 'edward-kirton',
              'status': 'Installed'}]

        .. versionchanged:: 0.4.1
            Changed method name from ``get_tools`` to ``get_repositories`` to
            better align with the Tool Shed concepts

        .. seealso:: bioblend.galaxy.tools.get_tool_panel()
        """
        return self._get()

    def show_repository(self, toolShed_id: str) -> Dict[str, Any]:
        """
        Get details of a given Tool Shed repository as it is installed on this
        Galaxy instance.

        :type toolShed_id: str
        :param toolShed_id: Encoded Tool Shed ID

        :rtype: dict
        :return: Information about the tool
          For example::

            {'changeset_revision': 'b17455fb6222',
             'ctx_rev': '8',
             'owner': 'aaron',
             'status': 'Installed',
             'url': '/api/tool_shed_repositories/82de4a4c7135b20a'}

        .. versionchanged:: 0.4.1
            Changed method name from ``show_tool`` to ``show_repository`` to
            better align with the Tool Shed concepts
        """
        return self._get(id=toolShed_id)

    def install_repository_revision(
        self,
        tool_shed_url: str,
        name: str,
        owner: str,
        changeset_revision: str,
        install_tool_dependencies: bool = False,
        install_repository_dependencies: bool = False,
        install_resolver_dependencies: bool = False,
        tool_panel_section_id: Optional[str] = None,
        new_tool_panel_section_label: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Install a specified repository revision from a specified Tool Shed into
        this Galaxy instance. This example demonstrates installation of a repository
        that contains valid tools, loading them into a section of the Galaxy tool
        panel or creating a new tool panel section.
        You can choose if tool dependencies or repository dependencies should be
        installed through the Tool Shed,
        (use ``install_tool_dependencies`` or ``install_repository_dependencies``)
        or through a resolver that supports installing dependencies
        (use ``install_resolver_dependencies``). Note that any combination of
        the three dependency resolving variables is valid.

        Installing the repository into an existing tool panel section requires
        the tool panel config file (e.g., tool_conf.xml, shed_tool_conf.xml, etc)
        to contain the given tool panel section:

            <section id="from_test_tool_shed" name="From Test Tool Shed" version="">
            </section>

        :type tool_shed_url: str
        :param tool_shed_url: URL of the Tool Shed from which the repository should
                              be installed from (e.g., ``https://testtoolshed.g2.bx.psu.edu``)

        :type name: str
        :param name: The name of the repository that should be installed

        :type owner: str
        :param owner: The name of the repository owner

        :type changeset_revision: str
        :param changeset_revision: The revision of the repository to be installed

        :type install_tool_dependencies: bool
        :param install_tool_dependencies: Whether or not to automatically handle
                                          tool dependencies (see
                                          https://galaxyproject.org/toolshed/tool-dependency-recipes/
                                          for more details)

        :type install_repository_dependencies: bool
        :param install_repository_dependencies: Whether or not to automatically
                                                handle repository dependencies
                                                (see https://galaxyproject.org/toolshed/defining-repository-dependencies/
                                                for more details)

        :type install_resolver_dependencies: bool
        :param install_resolver_dependencies: Whether or not to automatically
                                                install resolver dependencies (e.g. conda).

        :type tool_panel_section_id: str
        :param tool_panel_section_id: The ID of the Galaxy tool panel section
                                      where the tool should be insterted under.
                                      Note that you should specify either this
                                      parameter or the ``new_tool_panel_section_label``.
                                      If both are specified, this one will take
                                      precedence.

        :type new_tool_panel_section_label: str
        :param new_tool_panel_section_label: The name of a Galaxy tool panel section
                                             that should be created and the repository
                                             installed into.
        """
        payload: Dict[str, Any] = {}
        payload["tool_shed_url"] = tool_shed_url
        payload["name"] = name
        payload["owner"] = owner
        payload["changeset_revision"] = changeset_revision
        payload["install_tool_dependencies"] = install_tool_dependencies
        payload["install_repository_dependencies"] = install_repository_dependencies
        payload["install_resolver_dependencies"] = install_resolver_dependencies
        if tool_panel_section_id:
            payload["tool_panel_section_id"] = tool_panel_section_id
        elif new_tool_panel_section_label:
            payload["new_tool_panel_section_label"] = new_tool_panel_section_label

        url = self._make_url() + "/new/install_repository_revision"
        return self._post(url=url, payload=payload)

    def uninstall_repository_revision(
        self, name: str, owner: str, changeset_revision: str, tool_shed_url: str, remove_from_disk: bool = True
    ) -> Dict[str, Any]:
        """
        Uninstalls a specified repository revision from this Galaxy instance.

        :type name: str
        :param name: The name of the repository

        :type owner: str
        :param owner: The owner of the repository

        :type changeset_revision: str
        :param changeset_revision: The revision of the repository to uninstall

        :type tool_shed_url: str
        :param tool_shed_url: URL of the Tool Shed from which the repository was
          installed from (e.g., ``https://testtoolshed.g2.bx.psu.edu``)

        :type remove_from_disk: bool
        :param remove_from_disk: whether to also remove the repository from disk
          (the default) or only deactivate it

        :rtype: dict
        :return: If successful, a dictionary with a message noting the removal
        """
        payload: Dict[str, Any] = {
            "tool_shed_url": tool_shed_url,
            "name": name,
            "owner": owner,
            "changeset_revision": changeset_revision,
            "remove_from_disk": remove_from_disk,
        }
        return self._delete(params=payload)



"""
Contains possible interactions with the Galaxy Workflows
"""

import json
import os
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance

InputsBy = Literal["step_index|step_uuid", "step_index", "step_id", "step_uuid", "name"]  # type: ignore[name-defined]


class WorkflowClient(Client):
    module = "workflows"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    # the 'deleted' option is not available for workflows
    def get_workflows(
        self, workflow_id: Optional[str] = None, name: Optional[str] = None, published: bool = False
    ) -> List[Dict[str, Any]]:
        """
        Get all workflows, or select a subset by specifying optional arguments
        for filtering (e.g. a workflow name).

        :type name: str
        :param name: Workflow name to filter on.

        :type published: bool
        :param published: if ``True``, return also published workflows

        :rtype: list
        :return: A list of workflow dicts.
                 For example::

                   [{'id': '92c56938c2f9b315',
                     'name': 'Simple',
                     'url': '/api/workflows/92c56938c2f9b315'}]

        .. versionchanged:: 1.1.1
           Using the deprecated ``workflow_id`` parameter now raises a
           ``ValueError`` exception.
        """
        if workflow_id is not None:
            raise ValueError(
                "The workflow_id parameter has been removed, use the show_workflow() method to view details of a workflow for which you know the ID."
            )
        params: Dict[str, Any] = {}
        if published:
            params["show_published"] = True
        workflows = self._get(params=params)
        if name is not None:
            workflows = [_ for _ in workflows if _["name"] == name]
        return workflows

    def show_workflow(self, workflow_id: str, version: Optional[int] = None) -> Dict[str, Any]:
        """
        Display information needed to run a workflow.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type version: int
        :param version: Workflow version to show

        :rtype: dict
        :return: A description of the workflow and its inputs.
          For example::

            {'id': '92c56938c2f9b315',
             'inputs': {'23': {'label': 'Input Dataset', 'value': ''}},
             'name': 'Simple',
             'url': '/api/workflows/92c56938c2f9b315'}
        """
        params: Dict[str, Any] = {}
        if version is not None:
            params["version"] = version

        return self._get(id=workflow_id, params=params)

    def get_workflow_inputs(self, workflow_id: str, label: str) -> List[str]:
        """
        Get a list of workflow input IDs that match the given label.
        If no input matches the given label, an empty list is returned.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type label: str
        :param label: label to filter workflow inputs on

        :rtype: list
        :return: list of workflow inputs matching the label query
        """
        wf = self._get(id=workflow_id)
        inputs = wf["inputs"]
        return [id for id in inputs if inputs[id]["label"] == label]

    def import_workflow_dict(self, workflow_dict: Dict[str, Any], publish: bool = False) -> Dict[str, Any]:
        """
        Imports a new workflow given a dictionary representing a previously
        exported workflow.

        :type workflow_dict: dict
        :param workflow_dict: dictionary representing the workflow to be imported

        :type publish: bool
        :param publish:  if ``True`` the uploaded workflow will be published;
                         otherwise it will be visible only by the user which uploads it (default)

        :rtype: dict
        :return: Information about the imported workflow.
          For example::

            {'name': 'Training: 16S rRNA sequencing with mothur: main tutorial',
             'tags': [],
             'deleted': false,
             'latest_workflow_uuid': '368c6165-ccbe-4945-8a3c-d27982206d66',
             'url': '/api/workflows/94bac0a90086bdcf',
             'number_of_steps': 44,
             'published': false,
             'owner': 'jane-doe',
             'model_class': 'StoredWorkflow',
             'id': '94bac0a90086bdcf'}
        """
        payload = {"workflow": workflow_dict, "publish": publish}

        url = self._make_url() + "/upload"
        return self._post(url=url, payload=payload)

    def import_workflow_from_local_path(self, file_local_path: str, publish: bool = False) -> Dict[str, Any]:
        """
        Imports a new workflow given the path to a file containing a previously
        exported workflow.

        :type file_local_path: str
        :param file_local_path: File to upload to the server for new workflow

        :type publish: bool
        :param publish:  if ``True`` the uploaded workflow will be published;
                         otherwise it will be visible only by the user which uploads it (default)

        :rtype: dict
        :return: Information about the imported workflow.
          For example::

            {'name': 'Training: 16S rRNA sequencing with mothur: main tutorial',
             'tags': [],
             'deleted': false,
             'latest_workflow_uuid': '368c6165-ccbe-4945-8a3c-d27982206d66',
             'url': '/api/workflows/94bac0a90086bdcf',
             'number_of_steps': 44,
             'published': false,
             'owner': 'jane-doe',
             'model_class': 'StoredWorkflow',
             'id': '94bac0a90086bdcf'}

        """
        with open(file_local_path) as fp:
            workflow_json = json.load(fp)

        return self.import_workflow_dict(workflow_json, publish)

    def import_shared_workflow(self, workflow_id: str) -> Dict[str, Any]:
        """
        Imports a new workflow from the shared published workflows.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :rtype: dict
        :return: A description of the workflow.
          For example::

            {'id': 'ee0e2b4b696d9092',
             'model_class': 'StoredWorkflow',
             'name': 'Super workflow that solves everything!',
             'published': False,
             'tags': [],
             'url': '/api/workflows/ee0e2b4b696d9092'}
        """
        payload = {"shared_workflow_id": workflow_id}
        url = self._make_url()
        return self._post(url=url, payload=payload)

    def export_workflow_dict(self, workflow_id: str, version: Optional[int] = None) -> Dict[str, Any]:
        """
        Exports a workflow.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type version: int
        :param version: Workflow version to export

        :rtype: dict
        :return: Dictionary representing the requested workflow
        """
        params: Dict[str, Any] = {}
        if version is not None:
            params["version"] = version

        url = "/".join((self._make_url(), "download", workflow_id))
        return self._get(url=url, params=params)

    def export_workflow_to_local_path(
        self, workflow_id: str, file_local_path: str, use_default_filename: bool = True
    ) -> None:
        """
        Exports a workflow in JSON format to a given local path.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type file_local_path: str
        :param file_local_path: Local path to which the exported file will be saved.
                                (Should not contain filename if use_default_name=True)

        :type use_default_filename: bool
        :param use_default_filename: If the use_default_name parameter is True, the exported
          file will be saved as file_local_path/Galaxy-Workflow-%s.ga, where %s
          is the workflow name. If use_default_name is False, file_local_path
          is assumed to contain the full file path including filename.

        :rtype: None
        :return: None
        """
        workflow_dict = self.export_workflow_dict(workflow_id)

        if use_default_filename:
            filename = f"Galaxy-Workflow-{workflow_dict['name']}.ga"
            file_local_path = os.path.join(file_local_path, filename)

        with open(file_local_path, "w") as fp:
            json.dump(workflow_dict, fp)

    def update_workflow(self, workflow_id: str, **kwargs: Any) -> Dict[str, Any]:
        """
        Update a given workflow.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type workflow: dict
        :param workflow: dictionary representing the workflow to be updated

        :type name: str
        :param name: New name of the workflow

        :type annotation: str
        :param annotation: New annotation for the workflow

        :type menu_entry: bool
        :param menu_entry: Whether the workflow should appear in the user's menu

        :type tags: list of str
        :param tags: Replace workflow tags with the given list

        :type published: bool
        :param published: Whether the workflow should be published or unpublished

        :rtype: dict
        :return: Dictionary representing the updated workflow
        """
        return self._put(payload=kwargs, id=workflow_id)

    def invoke_workflow(
        self,
        workflow_id: str,
        inputs: Optional[dict] = None,
        params: Optional[dict] = None,
        history_id: Optional[str] = None,
        history_name: Optional[str] = None,
        import_inputs_to_history: bool = False,
        replacement_params: Optional[dict] = None,
        allow_tool_state_corrections: bool = False,
        inputs_by: Optional[InputsBy] = None,
        parameters_normalized: bool = False,
        require_exact_tool_versions: bool = True,
    ) -> Dict[str, Any]:
        """
        Invoke the workflow identified by ``workflow_id``. This will
        cause a workflow to be scheduled and return an object describing
        the workflow invocation.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type inputs: dict
        :param inputs: A mapping of workflow inputs to datasets and dataset collections.
                       The datasets source can be a LibraryDatasetDatasetAssociation (``ldda``),
                       LibraryDataset (``ld``), HistoryDatasetAssociation (``hda``), or
                       HistoryDatasetCollectionAssociation (``hdca``).

                       The map must be in the following format:
                       ``{'<input_index>': {'id': <encoded dataset ID>, 'src': '[ldda, ld, hda, hdca]'}}``
                       (e.g. ``{'2': {'id': '29beef4fadeed09f', 'src': 'hda'}}``)

                       This map may also be indexed by the UUIDs of the workflow steps,
                       as indicated by the ``uuid`` property of steps returned from the
                       Galaxy API. Alternatively workflow steps may be addressed by
                       the label that can be set in the workflow editor. If using
                       uuid or label you need to also set the ``inputs_by`` parameter
                       to ``step_uuid`` or ``name``.

        :type params: dict
        :param params: A mapping of non-datasets tool parameters (see below)

        :type history_id: str
        :param history_id: The encoded history ID where to store the workflow
          output. Alternatively, ``history_name`` may be specified to create a
          new history.

        :type history_name: str
        :param history_name: Create a new history with the given name to store
          the workflow output. If both ``history_id`` and ``history_name`` are
          provided, ``history_name`` is ignored. If neither is specified, a new
          'Unnamed history' is created.

        :type import_inputs_to_history: bool
        :param import_inputs_to_history: If ``True``, used workflow inputs will
          be imported into the history. If ``False``, only workflow outputs will
          be visible in the given history.

        :type allow_tool_state_corrections: bool
        :param allow_tool_state_corrections: If True, allow Galaxy to fill in
          missing tool state when running workflows. This may be useful for
          workflows using tools that have changed over time or for workflows
          built outside of Galaxy with only a subset of inputs defined.

        :type replacement_params: dict
        :param replacement_params: pattern-based replacements for post-job
          actions (see below)

        :type inputs_by: str
        :param inputs_by: Determines how inputs are referenced. Can be
          "step_index|step_uuid" (default), "step_index", "step_id", "step_uuid", or "name".

        :type parameters_normalized: bool
        :param parameters_normalized: Whether Galaxy should normalize ``params``
          to ensure everything is referenced by a numeric step ID. Default is
          ``False``, but when setting ``params`` for a subworkflow, ``True`` is
          required.

        :type require_exact_tool_versions: bool
        :param require_exact_tool_versions: Whether invocation should fail if
          Galaxy does not have the exact tool versions. Default is ``True``.
          Parameter does not any effect for Galaxy versions < 22.05.

        :rtype: dict
        :return: A dict containing the workflow invocation describing the
          scheduling of the workflow. For example::

            {'history_id': '2f94e8ae9edff68a',
             'id': 'df7a1f0c02a5b08e',
             'inputs': {'0': {'id': 'a7db2fac67043c7e',
                              'src': 'hda',
                              'uuid': '7932ffe0-2340-4952-8857-dbaa50f1f46a'}},
             'model_class': 'WorkflowInvocation',
             'state': 'ready',
             'steps': [{'action': None,
                        'id': 'd413a19dec13d11e',
                        'job_id': None,
                        'model_class': 'WorkflowInvocationStep',
                        'order_index': 0,
                        'state': None,
                        'update_time': '2015-10-31T22:00:26',
                        'workflow_step_id': 'cbbbf59e8f08c98c',
                        'workflow_step_label': None,
                        'workflow_step_uuid': 'b81250fd-3278-4e6a-b269-56a1f01ef485'},
                       {'action': None,
                        'id': '2f94e8ae9edff68a',
                        'job_id': 'e89067bb68bee7a0',
                        'model_class': 'WorkflowInvocationStep',
                        'order_index': 1,
                        'state': 'new',
                        'update_time': '2015-10-31T22:00:26',
                        'workflow_step_id': '964b37715ec9bd22',
                        'workflow_step_label': None,
                        'workflow_step_uuid': 'e62440b8-e911-408b-b124-e05435d3125e'}],
             'update_time': '2015-10-31T22:00:26',
             'uuid': 'c8aa2b1c-801a-11e5-a9e5-8ca98228593c',
             'workflow_id': '03501d7626bd192f'}

        The ``params`` dict should be specified as follows::

          {STEP_ID: PARAM_DICT, ...}

        where PARAM_DICT is::

          {PARAM_NAME: VALUE, ...}

        For backwards compatibility, the following (deprecated) format is
        also supported for ``params``::

          {TOOL_ID: PARAM_DICT, ...}

        in which case PARAM_DICT affects all steps with the given tool id.
        If both by-tool-id and by-step-id specifications are used, the
        latter takes precedence.

        Finally (again, for backwards compatibility), PARAM_DICT can also
        be specified as::

          {'param': PARAM_NAME, 'value': VALUE}

        Note that this format allows only one parameter to be set per step.

        For a ``repeat`` parameter, the names of the contained parameters needs
        to be specified as ``<repeat name>_<repeat index>|<param name>``, with
        the repeat index starting at 0. For example, if the tool XML contains::

          <repeat name="cutoff" title="Parameters used to filter cells" min="1">
              <param name="name" type="text" value="n_genes" label="Name of param...">
                  <option value="n_genes">n_genes</option>
                  <option value="n_counts">n_counts</option>
              </param>
              <param name="min" type="float" min="0" value="0" label="Min value"/>
          </repeat>

        then the PARAM_DICT should be something like::

          {...
           "cutoff_0|name": "n_genes",
           "cutoff_0|min": "2",
           "cutoff_1|name": "n_counts",
           "cutoff_1|min": "4",
           ...}

        At the time of this writing, it is not possible to change the number of
        times the contained parameters are repeated. Therefore, the parameter
        indexes can go from 0 to n-1, where n is the number of times the
        repeated element was added when the workflow was saved in the Galaxy UI.

        The ``replacement_params`` dict should map parameter names in
        post-job actions (PJAs) to their runtime values. For
        instance, if the final step has a PJA like the following::

          {'RenameDatasetActionout_file1': {'action_arguments': {'newname': '${output}'},
                                            'action_type': 'RenameDatasetAction',
                                            'output_name': 'out_file1'}}

        then the following renames the output dataset to 'foo'::

          replacement_params = {'output': 'foo'}

        see also `this email thread
        <http://lists.bx.psu.edu/pipermail/galaxy-dev/2011-September/006875.html>`_.

        .. warning::
          Historically, workflow invocation consumed a ``dataset_map``
          data structure that was indexed by unencoded workflow step IDs. These
          IDs would not be stable across Galaxy instances. The new ``inputs``
          property is instead indexed by either the ``order_index`` property
          (which is stable across workflow imports) or the step UUID which is
          also stable.
        """
        payload: Dict[str, Any] = {}
        if inputs:
            payload["inputs"] = inputs

        if params:
            payload["parameters"] = params

        if replacement_params:
            payload["replacement_params"] = replacement_params

        if history_id:
            payload["history"] = f"hist_id={history_id}"
        elif history_name:
            payload["history"] = history_name
        if not import_inputs_to_history:
            payload["no_add_to_history"] = True
        if allow_tool_state_corrections:
            payload["allow_tool_state_corrections"] = allow_tool_state_corrections
        if inputs_by is not None:
            payload["inputs_by"] = inputs_by
        payload["require_exact_tool_versions"] = require_exact_tool_versions
        if parameters_normalized:
            payload["parameters_normalized"] = parameters_normalized
        url = self._invocations_url(workflow_id)
        return self._post(payload, url=url)

    def show_invocation(self, workflow_id: str, invocation_id: str) -> Dict[str, Any]:
        """
        Get a workflow invocation object representing the scheduling of a
        workflow. This object may be sparse at first (missing inputs and
        invocation steps) and will become more populated as the workflow is
        actually scheduled.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :rtype: dict
        :return: The workflow invocation.
          For example::

            {'history_id': '2f94e8ae9edff68a',
             'id': 'df7a1f0c02a5b08e',
             'inputs': {'0': {'id': 'a7db2fac67043c7e',
                              'src': 'hda',
                              'uuid': '7932ffe0-2340-4952-8857-dbaa50f1f46a'}},
             'model_class': 'WorkflowInvocation',
             'state': 'ready',
             'steps': [{'action': None,
                        'id': 'd413a19dec13d11e',
                        'job_id': None,
                        'model_class': 'WorkflowInvocationStep',
                        'order_index': 0,
                        'state': None,
                        'update_time': '2015-10-31T22:00:26',
                        'workflow_step_id': 'cbbbf59e8f08c98c',
                        'workflow_step_label': None,
                        'workflow_step_uuid': 'b81250fd-3278-4e6a-b269-56a1f01ef485'},
                       {'action': None,
                        'id': '2f94e8ae9edff68a',
                        'job_id': 'e89067bb68bee7a0',
                        'model_class': 'WorkflowInvocationStep',
                        'order_index': 1,
                        'state': 'new',
                        'update_time': '2015-10-31T22:00:26',
                        'workflow_step_id': '964b37715ec9bd22',
                        'workflow_step_label': None,
                        'workflow_step_uuid': 'e62440b8-e911-408b-b124-e05435d3125e'}],
             'update_time': '2015-10-31T22:00:26',
             'uuid': 'c8aa2b1c-801a-11e5-a9e5-8ca98228593c',
             'workflow_id': '03501d7626bd192f'}
        """
        url = self._invocation_url(workflow_id, invocation_id)
        return self._get(url=url)

    def get_invocations(self, workflow_id: str) -> List[Dict[str, Any]]:
        """
        Get a list containing all the workflow invocations corresponding to the
        specified workflow.

        For more advanced filtering use InvocationClient.get_invocations().

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :rtype: list
        :return: A list of workflow invocations.
          For example::

            [{'history_id': '2f94e8ae9edff68a',
              'id': 'df7a1f0c02a5b08e',
              'model_class': 'WorkflowInvocation',
              'state': 'new',
              'update_time': '2015-10-31T22:00:22',
              'uuid': 'c8aa2b1c-801a-11e5-a9e5-8ca98228593c',
              'workflow_id': '03501d7626bd192f'}]
        """
        url = self._invocations_url(workflow_id)
        return self._get(url=url)

    def cancel_invocation(self, workflow_id: str, invocation_id: str) -> Dict[str, Any]:
        """
        Cancel the scheduling of a workflow.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :rtype: dict
        :return: The workflow invocation being cancelled
        """
        url = self._invocation_url(workflow_id, invocation_id)
        return self._delete(url=url)

    def show_invocation_step(self, workflow_id: str, invocation_id: str, step_id: str) -> Dict[str, Any]:
        """
        See the details of a particular workflow invocation step.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :type step_id: str
        :param step_id: Encoded workflow invocation step ID

        :rtype: dict
        :return: The workflow invocation step.
          For example::

            {'action': None,
             'id': '63cd3858d057a6d1',
             'job_id': None,
             'model_class': 'WorkflowInvocationStep',
             'order_index': 2,
             'state': None,
             'update_time': '2015-10-31T22:11:14',
             'workflow_step_id': '52e496b945151ee8',
             'workflow_step_label': None,
             'workflow_step_uuid': '4060554c-1dd5-4287-9040-8b4f281cf9dc'}
        """
        url = self._invocation_step_url(workflow_id, invocation_id, step_id)
        return self._get(url=url)

    def run_invocation_step_action(
        self, workflow_id: str, invocation_id: str, step_id: str, action: Any
    ) -> Dict[str, Any]:
        """Execute an action for an active workflow invocation step. The
        nature of this action and what is expected will vary based on the
        the type of workflow step (the only currently valid action is True/False
        for pause steps).

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type invocation_id: str
        :param invocation_id: Encoded workflow invocation ID

        :type step_id: str
        :param step_id: Encoded workflow invocation step ID

        :type action: object
        :param action: Action to use when updating state, semantics depends on
           step type.

        :rtype: dict
        :return: Representation of the workflow invocation step
        """
        url = self._invocation_step_url(workflow_id, invocation_id, step_id)
        payload = {"action": action}
        return self._put(payload=payload, url=url)

    def delete_workflow(self, workflow_id: str) -> None:
        """
        Delete a workflow identified by `workflow_id`.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        .. warning::
            Deleting a workflow is irreversible in Galaxy versions < 23.01 - all
            workflow data will be permanently deleted.
        """
        self._delete(id=workflow_id)

    def refactor_workflow(
        self, workflow_id: str, actions: List[Dict[str, Any]], dry_run: bool = False
    ) -> Dict[str, Any]:
        """
        Refactor workflow with given actions.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :type actions: list of dicts
        :param actions: Actions to use for refactoring the workflow. The following
                        actions are supported: update_step_label, update_step_position,
                        update_output_label, update_name, update_annotation,
                        update_license, update_creator, update_report, add_step,
                        add_input, disconnect, connect, fill_defaults, fill_step_defaults,
                        extract_input, extract_legacy_parameter,
                        remove_unlabeled_workflow_outputs, upgrade_all_steps,
                        upgrade_subworkflow, upgrade_tool.

          An example value for the ``actions`` argument might be::

            actions = [
                {"action_type": "add_input", "type": "data", "label": "foo"},
                {"action_type": "update_step_label", "label": "bar", "step": {"label": "foo"}},
            ]

        :type dry_run: bool
        :param dry_run: When true, perform a dry run where the existing
                        workflow is preserved. The refactored workflow
                        is returned in the output of the method, but not saved
                        on the Galaxy server.

        :rtype: dict
        :return: Dictionary containing logged messages for the executed actions
                 and the refactored workflow.
        """
        payload = {
            "actions": actions,
            "dry_run": dry_run,
        }
        url = "/".join((self._make_url(workflow_id), "refactor"))
        return self._put(payload=payload, url=url)

    def extract_workflow_from_history(
        self,
        history_id: str,
        workflow_name: str,
        job_ids: Optional[List[str]] = None,
        dataset_hids: Optional[List[str]] = None,
        dataset_collection_hids: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Extract a workflow from a history.

        :type history_id: str
        :param history_id: Encoded history ID

        :type   workflow_name: str
        :param  workflow_name: Name of the workflow to create

        :type   job_ids: list
        :param  job_ids: Optional list of job IDs to filter the jobs to extract from the history

        :type   dataset_hids: list
        :param  dataset_hids: Optional list of dataset hids corresponding to workflow inputs
                             when extracting a workflow from history

        :type   dataset_collection_hids: list
        :param  dataset_collection_hids: Optional list of dataset collection hids corresponding to workflow inputs
                                        when extracting a workflow from history

        :rtype: dict
        :return: A description of the created workflow
        """
        payload = {
            "from_history_id": history_id,
            "job_ids": job_ids if job_ids else [],
            "dataset_ids": dataset_hids if dataset_hids else [],
            "dataset_collection_ids": dataset_collection_hids if dataset_collection_hids else [],
            "workflow_name": workflow_name,
        }
        return self._post(payload=payload)

    def show_versions(self, workflow_id: str) -> List[Dict[str, Any]]:
        """
        Get versions for a workflow.

        :type workflow_id: str
        :param workflow_id: Encoded workflow ID

        :rtype: list of dicts
        :return: Ordered list of version descriptions for this workflow
        """
        url = self._make_url(workflow_id) + "/versions"
        return self._get(url=url)

    def _invocation_step_url(self, workflow_id: str, invocation_id: str, step_id: str) -> str:
        return "/".join((self._invocation_url(workflow_id, invocation_id), "steps", step_id))

    def _invocation_url(self, workflow_id: str, invocation_id: str) -> str:
        return "/".join((self._invocations_url(workflow_id), invocation_id))

    def _invocations_url(self, workflow_id: str) -> str:
        return "/".join((self._make_url(workflow_id), "invocations"))


__all__ = ("WorkflowClient",)



"""
Interaction with a Tool Shed instance categories
"""

from typing import (
    Any,
    Dict,
    List,
    Literal,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.toolshed import ToolShedInstance


class ToolShedCategoryClient(Client):
    module = "categories"

    def __init__(self, toolshed_instance: "ToolShedInstance") -> None:
        super().__init__(toolshed_instance)

    def get_categories(self, deleted: bool = False) -> List[Dict[str, Any]]:
        """
        Returns a list of dictionaries that contain descriptions of the
        repository categories found on the given Tool Shed instance.

        :type deleted: bool
        :param deleted: whether to show deleted categories. Requires
          administrator access to the Tool Shed instance.

        :rtype: list
        :return: A list of dictionaries containing information about
          repository categories present in the Tool Shed.
          For example::

            [{'deleted': False,
              'description': 'Tools for manipulating data',
              'id': '175812cd7caaf439',
              'model_class': 'Category',
              'name': 'Text Manipulation',
              'url': '/api/categories/175812cd7caaf439'}]

        .. versionadded:: 0.5.2
        """
        return self._get(deleted=deleted)

    def show_category(self, category_id: str) -> Dict[str, Any]:
        """
        Get details of a given category.

        :type category_id: str
        :param category_id: Encoded category ID

        :rtype: dict
        :return: details of the given category
        """
        return self._get(id=category_id)

    def get_repositories(
        self, category_id: str, sort_key: Literal["name", "owner"] = "name", sort_order: Literal["asc", "desc"] = "asc"
    ) -> Dict[str, Any]:
        """
        Returns a dictionary of information for a repository category including
        a list of repositories belonging to the category.

        :type category_id: str
        :param category_id: Encoded category ID

        :type  sort_key: str
        :param sort_key: key for sorting. Options are 'name' or 'owner' (default 'name').

        :type  sort_order: str
        :param sort_order: ordering of sorted output. Options are 'asc' or 'desc' (default 'asc').

        :rtype: dict
        :return: A dict containing information about the category
          including a list of repository dicts.
          For example::

            {'deleted': False,
             'description': 'Tools for constructing and analyzing 3-dimensional shapes and '
                            'their properties',
             'id': '589548af7e391bcf',
             'model_class': 'Category',
             'name': 'Constructive Solid Geometry',
             'repositories': [{'create_time': '2016-08-23T18:53:23.845013',
                               'deleted': False,
                               'deprecated': False,
                               'description': 'Adds a surface field to a selected shape '
                                              'based on a given mathematical expression',
                               'homepage_url': 'https://github.com/gregvonkuster/galaxy-csg',
                               'id': 'af2ccc53697b064c',
                               'metadata': {'0:e12b55e960de': {'changeset_revision': 'e12b55e960de',
                                                               'downloadable': True,
                                                               'has_repository_dependencies': False,
                                                               'id': 'dfe022067783215f',
                                                               'includes_datatypes': False,
                                                               'includes_tool_dependencies': False,
                                                               'includes_tools': True,
                                                               'includes_tools_for_display_in_tool_panel': True,
                                                               'includes_workflows': False,
                                                               'malicious': False,
                                                               'missing_test_components': False,
                                                               'model_class': 'RepositoryMetadata',
                                                               'numeric_revision': 0,
                                                               'repository_id': 'af2ccc53697b064c'}},
                               'model_class': 'Repository',
                               'name': 'icqsol_add_surface_field_from_expression',
                               'owner': 'iuc',
                               'private': False,
                               'remote_repository_url': 'https://github.com/gregvonkuster/galaxy-csg',
                               'times_downloaded': 152,
                               'type': 'unrestricted',
                               'user_id': 'b563abc230aa8fd0'},
                              # ...
                              ],
             'repository_count': 11,
             'url': '/api/categories/589548af7e391bcf'}
        """

        params: Dict[str, Any] = {}
        if sort_key:
            params.update({"sort_key": sort_key})
        if sort_order:
            params.update({"sort_order": sort_order})

        url = self._make_url(category_id) + "/repositories"
        return self._get(url=url, params=params)



import os
import tempfile
from typing import (
    Any,
    List,
)

from . import (
    GalaxyTestBase,
    test_util,
)

FOO_DATA = "foo\nbar\n"


def listify(item: Any) -> List:
    # Slightly simplified version of listify() from
    # https://github.com/galaxyproject/galaxy/blob/dev/lib/galaxy/util/__init__.py
    if not item:
        return []
    elif isinstance(item, (list, tuple)):
        return list(item)
    elif isinstance(item, str) and item.count(","):
        return [token.strip() for token in item.split(",")]
    return [item]


class TestGalaxyLibraries(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        self.name = "automated test library"
        self.library = self.gi.libraries.create_library(
            self.name, description="automated test", synopsis="automated test synopsis"
        )

    def tearDown(self):
        self.gi.libraries.delete_library(self.library["id"])

    def test_create_library(self):
        assert self.library["name"] == self.name
        assert self.library["id"] is not None

    def test_get_libraries(self):
        libraries_with_name = self.gi.libraries.get_libraries(name=self.name)
        assert len([lib for lib in libraries_with_name if lib["id"] == self.library["id"]]) == 1

        deleted_name = "deleted test library"
        deleted_library_id = self.gi.libraries.create_library(
            deleted_name, description="a deleted library", synopsis="automated test synopsis"
        )["id"]
        self.gi.libraries.delete_library(deleted_library_id)
        deleted_libraries_with_name = self.gi.libraries.get_libraries(name=deleted_name, deleted=True)
        assert len([lib for lib in deleted_libraries_with_name if lib["id"] == deleted_library_id]) == 1

        all_non_deleted_libraries = self.gi.libraries.get_libraries(deleted=False)
        assert len([lib for lib in all_non_deleted_libraries if lib["id"] == self.library["id"]]) == 1
        assert [lib for lib in all_non_deleted_libraries if lib["id"] == deleted_library_id] == []

        all_deleted_libraries = self.gi.libraries.get_libraries(deleted=True)
        assert [lib for lib in all_deleted_libraries if lib["id"] == self.library["id"]] == []
        assert len([lib for lib in all_deleted_libraries if lib["id"] == deleted_library_id]) == 1

        all_libraries = self.gi.libraries.get_libraries(deleted=None)
        assert len([lib for lib in all_libraries if lib["id"] == self.library["id"]]) == 1
        assert len([lib for lib in all_libraries if lib["id"] == deleted_library_id]) == 1

    def test_show_library(self):
        library_data = self.gi.libraries.show_library(self.library["id"])
        assert self.library["id"] == library_data["id"]
        assert self.library["name"] == library_data["name"]

    def test_upload_file_from_url(self):
        url = "https://zenodo.org/record/582600/files/wildtype.fna?download=1"
        ret = self.gi.libraries.upload_file_from_url(self.library["id"], url)
        assert len(ret) == 1
        ldda_dict = ret[0]
        assert ldda_dict["name"] == url

    def test_upload_file_contents(self):
        ret = self.gi.libraries.upload_file_contents(self.library["id"], FOO_DATA)
        assert len(ret) == 1
        ldda_dict = ret[0]
        assert ldda_dict["name"] == "Pasted Entry"

    def test_upload_file_from_local_path(self):
        with tempfile.NamedTemporaryFile(mode="w", prefix="bioblend_test_") as f:
            f.write(FOO_DATA)
            f.flush()
            filename = f.name
            ret = self.gi.libraries.upload_file_from_local_path(self.library["id"], filename)
        assert len(ret) == 1
        ldda_dict = ret[0]
        assert ldda_dict["name"] == os.path.basename(filename)

    # def test_upload_file_from_server(self):
    #     pass

    def test_upload_from_galaxy_filesystem(self):
        bnames = [f"f{i}.txt" for i in range(2)]
        with tempfile.TemporaryDirectory() as tempdir:
            fnames = [os.path.join(tempdir, _) for _ in bnames]
            for fn in fnames:
                with open(fn, "w") as f:
                    f.write(FOO_DATA)
            filesystem_paths = "\n".join(fnames)
            ret = self.gi.libraries.upload_from_galaxy_filesystem(self.library["id"], filesystem_paths)
            for fn, dataset_dict in zip(fnames, ret):
                dataset = self.gi.libraries.wait_for_dataset(self.library["id"], dataset_dict["id"])
                assert dataset["state"] == "ok"
                assert dataset["name"] == os.path.basename(fn)
            ret = self.gi.libraries.upload_from_galaxy_filesystem(
                self.library["id"], filesystem_paths, link_data_only="link_to_files"
            )
            for fn, dataset_dict in zip(fnames, ret):
                dataset = self.gi.libraries.wait_for_dataset(self.library["id"], dataset_dict["id"])
                assert dataset["state"] == "ok"
                assert dataset["name"] == os.path.basename(fn)

    def test_copy_from_dataset(self):
        history = self.gi.histories.create_history()
        dataset_id = self._test_dataset(history["id"])
        self.gi.libraries.copy_from_dataset(self.library["id"], dataset_id, message="Copied from dataset")

    def test_update_dataset(self):
        library_id = self.library["id"]
        dataset1 = self.gi.libraries.upload_file_contents(library_id, FOO_DATA)
        updated_dataset = self.gi.libraries.update_library_dataset(
            dataset1[0]["id"], name="Modified name", misc_info="Modified the name succesfully"
        )
        assert updated_dataset["name"] == "Modified name"
        assert updated_dataset["misc_info"] == "Modified the name succesfully"

    def test_library_permissions(self):
        current_user = self.gi.users.get_current_user()
        user_id_list_new = [current_user["id"]]
        self.gi.libraries.set_library_permissions(
            self.library["id"],
            access_in=user_id_list_new,
            modify_in=user_id_list_new,
            add_in=user_id_list_new,
            manage_in=user_id_list_new,
        )
        ret = self.gi.libraries.get_library_permissions(self.library["id"])
        assert {_[1] for _ in ret["access_library_role_list"]} == set(user_id_list_new)
        assert {_[1] for _ in ret["modify_library_role_list"]} == set(user_id_list_new)
        assert {_[1] for _ in ret["add_library_item_role_list"]} == set(user_id_list_new)
        assert {_[1] for _ in ret["manage_library_role_list"]} == set(user_id_list_new)

    def test_dataset_permissions(self):
        current_user = self.gi.users.get_current_user()
        user_id_list_new = [current_user["id"]]
        library_id = self.library["id"]
        dataset1 = self.gi.libraries.upload_file_contents(library_id, FOO_DATA)
        ret = self.gi.libraries.set_dataset_permissions(
            dataset1[0]["id"], access_in=user_id_list_new, modify_in=user_id_list_new, manage_in=user_id_list_new
        )
        assert {_[1] for _ in ret["access_dataset_roles"]} == set(user_id_list_new)
        assert {_[1] for _ in ret["modify_item_roles"]} == set(user_id_list_new)
        assert {_[1] for _ in ret["manage_dataset_roles"]} == set(user_id_list_new)
        # test get_dataset_permissions
        ret_get = self.gi.libraries.get_dataset_permissions(dataset1[0]["id"])
        assert {_[1] for _ in ret_get["access_dataset_roles"]} == set(user_id_list_new)
        assert {_[1] for _ in ret_get["modify_item_roles"]} == set(user_id_list_new)
        assert {_[1] for _ in ret_get["manage_dataset_roles"]} == set(user_id_list_new)

    @test_util.skip_unless_galaxy("release_19.09")
    def test_upload_file_contents_with_tags(self):
        datasets = self.gi.libraries.upload_file_contents(self.library["id"], FOO_DATA, tags=["name:foobar", "barfoo"])
        dataset_show = self.gi.libraries.show_dataset(self.library["id"], datasets[0]["id"])
        assert listify(dataset_show["tags"]) == ["name:foobar", "barfoo"]

    @test_util.skip_unless_galaxy("release_19.09")
    def test_update_dataset_tags(self):
        datasets = self.gi.libraries.upload_file_contents(self.library["id"], FOO_DATA)
        dataset_show = self.gi.libraries.show_dataset(self.library["id"], datasets[0]["id"])
        assert listify(dataset_show["tags"]) == []

        updated_dataset = self.gi.libraries.update_library_dataset(datasets[0]["id"], tags=["name:foobar", "barfoo"])
        dataset_show = self.gi.libraries.show_dataset(self.library["id"], updated_dataset["id"])

        assert listify(dataset_show["tags"]) == ["name:foobar", "barfoo"]



from bioblend.galaxy.tools.inputs import (
    conditional,
    dataset,
    inputs,
    repeat,
)


def test_conditional():
    # Build up example inputs for random_lines1
    as_dict = (
        inputs()
        .set("num_lines", 5)
        .set("input", dataset("encoded1"))
        .set("seed_source", conditional().set("seed_source_selector", "set_seed").set("seed", "asdf"))
        .to_dict()
    )
    assert as_dict["num_lines"] == 5
    assert as_dict["input"]["src"] == "hda"
    assert as_dict["input"]["id"] == "encoded1"
    assert as_dict["seed_source|seed_source_selector"] == "set_seed"
    assert as_dict["seed_source|seed"] == "asdf"


def test_repeat():
    # Build up inputs for cat1
    as_dict = (
        inputs()
        .set("input1", dataset("encoded1"))
        .set(
            "queries",
            repeat()
            .instance(inputs().set_dataset_param("input2", "encoded2"))
            .instance(inputs().set_dataset_param("input2", "encoded3")),
        )
        .to_dict()
    )
    assert as_dict["input1"]["src"] == "hda"
    assert as_dict["input1"]["id"] == "encoded1"
    assert as_dict["queries_0|input2"]["src"] == "hda"
    assert as_dict["queries_0|input2"]["id"] == "encoded2"
    assert as_dict["queries_1|input2"]["src"] == "hda"
    assert as_dict["queries_1|input2"]["id"] == "encoded3"



import configparser
import os
from typing import (
    IO,
    Optional,
)

BioBlendConfigPath = "/etc/bioblend.cfg"
BioBlendConfigLocations = [BioBlendConfigPath]
UserConfigPath = os.path.join(os.path.expanduser("~"), ".bioblend")
BioBlendConfigLocations.append(UserConfigPath)


class Config(configparser.ConfigParser):
    """
    BioBlend allows library-wide configuration to be set in external files.
    These configuration files can be used to specify access keys, for example.
    By default we use two locations for the BioBlend configurations:

    * System wide: ``/etc/bioblend.cfg``
    * Individual user: ``~/.bioblend`` (which works on both Windows and Unix)
    """

    def __init__(self, path: Optional[str] = None, fp: Optional[IO[str]] = None, do_load: bool = True) -> None:
        super().__init__({"working_dir": "/mnt/pyami", "debug": "0"})
        if do_load:
            if path:
                self.read([path])
            elif fp:
                self.read_file(fp)
            else:
                self.read(BioBlendConfigLocations)



from typing import (
    Dict,
    List,
)

from . import GalaxyTestBase

FOO_DATA = "foo\nbar\n"


class TestGalaxyFolders(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        self.name = "automated test folder"
        self.library = self.gi.libraries.create_library(
            self.name, description="automated test", synopsis="automated test synopsis"
        )
        self.folder = self.gi.folders.create_folder(
            self.library["root_folder_id"], self.name, description="automatically created folder"
        )

    def tearDown(self):
        self.gi.libraries.delete_library(self.library["id"])

    def test_create_folder(self):
        assert self.folder["name"] == self.name
        assert self.folder["description"] == "automatically created folder"

    def test_show_folder(self):
        f2 = self.gi.folders.show_folder(self.folder["id"])
        assert f2["id"] == self.folder["id"]

    def test_show_folder_contents(self):
        f2 = self.gi.folders.show_folder(self.folder["id"], contents=True)
        assert "folder_contents" in f2
        assert "metadata" in f2
        assert self.name == f2["metadata"]["folder_name"]

    def test_delete_folder(self):
        self.sub_folder = self.gi.folders.create_folder(self.folder["id"], self.name)
        self.gi.folders.delete_folder(self.sub_folder["id"])

    def test_update_folder(self):
        self.folder = self.gi.folders.update_folder(self.folder["id"], "new-name", "new-description")
        assert self.folder["name"] == "new-name"
        assert self.folder["description"] == "new-description"

    def test_get_set_permissions(self):
        empty_permission: Dict[str, List] = {
            "add_library_item_role_list": [],
            "modify_folder_role_list": [],
            "manage_folder_role_list": [],
        }
        # They should be empty to start with
        assert self.gi.folders.get_permissions(self.folder["id"], scope="current") == empty_permission
        assert self.gi.folders.get_permissions(self.folder["id"], scope="available") == empty_permission
        # Then we'll add a role
        role = self.gi.roles.get_roles()[0]
        self.gi.folders.set_permissions(self.folder["id"], add_ids=[role["id"]])
        assert (
            role["id"]
            in self.gi.folders.get_permissions(self.folder["id"], scope="available")["add_library_item_role_list"][0]
        )



import os
import unittest
from typing import (
    Any,
    Dict,
    Literal,
)

import bioblend
from bioblend.galaxy import GalaxyInstance
from . import test_util

bioblend.set_stream_logger("test", level="INFO")

BIOBLEND_TEST_JOB_TIMEOUT = int(os.environ.get("BIOBLEND_TEST_JOB_TIMEOUT", "60"))


@test_util.skip_unless_galaxy()
class GalaxyTestBase(unittest.TestCase):
    gi: GalaxyInstance

    @classmethod
    def setUpClass(cls):
        galaxy_key = os.environ["BIOBLEND_GALAXY_API_KEY"]
        galaxy_url = os.environ["BIOBLEND_GALAXY_URL"]
        cls.gi = GalaxyInstance(url=galaxy_url, key=galaxy_key)

    def _test_dataset(self, history_id: str, contents: str = "1\t2\t3", **kwargs: Any) -> str:
        tool_output = self.gi.tools.paste_content(contents, history_id, **kwargs)
        return tool_output["outputs"][0]["id"]

    def _wait_and_verify_dataset(
        self, dataset_id: str, expected_contents: bytes, timeout_seconds: float = BIOBLEND_TEST_JOB_TIMEOUT
    ) -> None:
        dataset_contents = self.gi.datasets.download_dataset(dataset_id, maxwait=timeout_seconds)
        assert dataset_contents == expected_contents

    def _run_random_lines1(
        self, history_id: str, dataset_id: str, input_format: Literal["21.01", "legacy"] = "legacy"
    ) -> Dict[str, Any]:
        tool_inputs = {
            "num_lines": "1",
            "input": {"src": "hda", "id": dataset_id},
        }
        if input_format == "21.01":
            tool_inputs.update({"seed_source": {"seed_source_selector": "set_seed", "seed": "asdf"}})
        else:
            # legacy format
            tool_inputs.update({"seed_source|seed_source_selector": "set_seed", "seed_source|seed": "asdf"})
        return self.gi.tools.run_tool(
            history_id=history_id, tool_id="random_lines1", tool_inputs=tool_inputs, input_format=input_format
        )



import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    TYPE_CHECKING,
    Union,
)

from bioblend import (
    CHUNK_SIZE,
    NotReady,
    TimeoutException,
    wait_on,
)
from bioblend.galaxy.client import Client
from bioblend.galaxy.datasets import TERMINAL_STATES

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance

log = logging.getLogger(__name__)


class HasElements:
    def __init__(
        self,
        name: str,
        type: str = "list",
        elements: Optional[Union[List[Union["CollectionElement", "SimpleElement"]], Dict[str, Any]]] = None,
    ) -> None:
        self.name = name
        self.type = type
        if isinstance(elements, dict):
            self.elements: List[Union["CollectionElement", "SimpleElement"]] = [
                HistoryDatasetElement(name=key, id=value) for key, value in elements.values()
            ]
        elif elements:
            self.elements = elements

    def add(self, element: Union["CollectionElement", "SimpleElement"]) -> "HasElements":
        self.elements.append(element)
        return self


class CollectionDescription(HasElements):
    def to_dict(self) -> Dict[str, Union[str, List]]:
        return {
            "name": self.name,
            "collection_type": self.type,
            "element_identifiers": [e.to_dict() for e in self.elements],
        }


class CollectionElement(HasElements):
    def to_dict(self) -> Dict[str, Union[str, List]]:
        return {
            "src": "new_collection",
            "name": self.name,
            "collection_type": self.type,
            "element_identifiers": [e.to_dict() for e in self.elements],
        }


class SimpleElement:
    def __init__(self, value: Dict[str, str]) -> None:
        self.value = value

    def to_dict(self) -> Dict[str, str]:
        return self.value


class HistoryDatasetElement(SimpleElement):
    def __init__(self, name: str, id: str) -> None:
        super().__init__(
            {
                "name": name,
                "src": "hda",
                "id": id,
            }
        )


class HistoryDatasetCollectionElement(SimpleElement):
    def __init__(self, name: str, id: str) -> None:
        super().__init__(
            {
                "name": name,
                "src": "hdca",
                "id": id,
            }
        )


class LibraryDatasetElement(SimpleElement):
    def __init__(self, name: str, id: str) -> None:
        super().__init__(
            {
                "name": name,
                "src": "ldda",
                "id": id,
            }
        )


class DatasetCollectionClient(Client):
    gi: "GalaxyInstance"
    module = "dataset_collections"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def show_dataset_collection(self, dataset_collection_id: str, instance_type: str = "history") -> Dict[str, Any]:
        """
        Get details of a given dataset collection of the current user

        :type dataset_collection_id: str
        :param dataset_collection_id: dataset collection ID

        :type instance_type: str
        :param instance_type: instance type of the collection - 'history' or 'library'

        :rtype: dict
        :return: element view of the dataset collection
        """
        params = {
            "instance_type": instance_type,
        }
        url = self._make_url(module_id=dataset_collection_id)
        return self._get(id=dataset_collection_id, url=url, params=params)

    def download_dataset_collection(self, dataset_collection_id: str, file_path: str) -> Dict[str, Any]:
        """
        Download a history dataset collection as an archive.

        :type dataset_collection_id: str
        :param dataset_collection_id: Encoded dataset collection ID

        :type file_path: str
        :param file_path: The path to which the archive will be downloaded

        :rtype: dict
        :return: Information about the downloaded archive.

        .. note::
          This method downloads a ``zip`` archive for Galaxy 21.01 and later.
          For earlier versions of Galaxy this method downloads a ``tgz`` archive.
        """
        url = self._make_url(module_id=dataset_collection_id) + "/download"
        r = self.gi.make_get_request(url, stream=True)
        r.raise_for_status()

        archive_type = "zip" if self.gi.config.get_version()["version_major"] >= "21.01" else "tgz"

        with open(file_path, "wb") as fp:
            for chunk in r.iter_content(chunk_size=CHUNK_SIZE):
                if chunk:
                    fp.write(chunk)

        return {"file_path": file_path, "archive_type": archive_type}

    def wait_for_dataset_collection(
        self,
        dataset_collection_id: str,
        maxwait: float = 12000,
        interval: float = 3,
        proportion_complete: float = 1.0,
        check: bool = True,
    ) -> Dict[str, Any]:
        """
        Wait until all or a specified proportion of elements of a dataset
        collection are in a terminal state.

        :type dataset_collection_id: str
        :param dataset_collection_id: dataset collection ID

        :type maxwait: float
        :param maxwait: Total time (in seconds) to wait for the dataset
          states in the dataset collection to become terminal. After this time,
          a ``TimeoutException`` will be raised.

        :type interval: float
        :param interval: Time (in seconds) to wait between two consecutive checks.

        :type proportion_complete: float
        :param proportion_complete: Proportion of elements in this collection
          that have to be in a terminal state for this method to return. Must
          be a number between 0 and 1. For example: if the dataset collection
          contains 2 elements, and proportion_complete=0.5 is specified, then
          wait_for_dataset_collection will return as soon as 1 of the 2
          datasets is in a terminal state. Default is 1, i.e. all elements must
          complete.

        :type check: bool
        :param check: Whether to check if all the terminal states of datasets
          in the dataset collection are 'ok'. This will raise an Exception if
          a dataset is in a terminal state other than 'ok'.

        :rtype: dict
        :return: Details of the given dataset collection.
        """
        assert 0 <= proportion_complete <= 1

        def check_and_get_dataset_collection() -> Dict[str, Any]:
            dataset_collection = self.show_dataset_collection(dataset_collection_id)
            states = [elem["object"]["state"] for elem in dataset_collection["elements"]]
            terminal_states = [state for state in states if state in TERMINAL_STATES]
            if set(terminal_states) not in [{"ok"}, set()]:
                raise Exception(
                    f"Dataset collection {dataset_collection_id} contains elements in the "
                    f"following non-ok terminal states: {', '.join(set(terminal_states) - {'ok'})}"
                )
            proportion = len(terminal_states) / len(states)
            if proportion >= proportion_complete:
                return dataset_collection
            raise NotReady(
                f"The dataset collection {dataset_collection_id} has only {proportion * 100}% of datasets in a terminal state"
            )

        return wait_on(check_and_get_dataset_collection, maxwait=maxwait, interval=interval)


# Unused, for backward compatibility
DatasetCollectionTimeoutException = TimeoutException


__all__ = (
    "CollectionDescription",
    "CollectionElement",
    "DatasetCollectionClient",
    "HistoryDatasetElement",
    "HistoryDatasetCollectionElement",
    "LibraryDatasetElement",
)



import json
import os
import shutil
import tempfile
import time
from typing import (
    Any,
    Dict,
    List,
)

import pytest

from bioblend import ConnectionError
from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyWorkflows(GalaxyTestBase.GalaxyTestBase):
    @test_util.skip_unless_tool("cat1")
    @test_util.skip_unless_tool("cat")
    def test_workflow_scheduling(self):
        path = test_util.get_abspath(os.path.join("data", "test_workflow_pause.ga"))
        workflow = self.gi.workflows.import_workflow_from_local_path(path)
        workflow_id = workflow["id"]
        history_id = self.gi.histories.create_history(name="TestWorkflowState")["id"]

        invocations = self.gi.workflows.get_invocations(workflow_id)
        assert len(invocations) == 0

        # Try invalid invocation (no input)
        with pytest.raises(ConnectionError):
            self.gi.workflows.invoke_workflow(workflow["id"])

        dataset1_id = self._test_dataset(history_id)
        invocation = self.gi.workflows.invoke_workflow(
            workflow["id"],
            inputs={"0": {"src": "hda", "id": dataset1_id}},
        )
        assert invocation["state"] == "new"
        invocation_id = invocation["id"]
        invocations = self.gi.workflows.get_invocations(workflow_id)
        assert len(invocations) == 1
        assert invocations[0]["id"] == invocation_id

        def invocation_steps_by_order_index() -> Dict[int, Dict[str, Any]]:
            invocation = self.gi.workflows.show_invocation(workflow_id, invocation_id)
            return {s["order_index"]: s for s in invocation["steps"]}

        for _ in range(20):
            if 2 in invocation_steps_by_order_index():
                break
            time.sleep(0.5)

        invocation = self.gi.workflows.show_invocation(workflow_id, invocation_id)
        assert invocation["state"] == "ready"

        steps = invocation_steps_by_order_index()
        pause_step = steps[2]
        assert self.gi.workflows.show_invocation_step(workflow_id, invocation_id, pause_step["id"])["action"] is None
        self.gi.workflows.run_invocation_step_action(workflow_id, invocation_id, pause_step["id"], action=True)
        assert self.gi.workflows.show_invocation_step(workflow_id, invocation_id, pause_step["id"])["action"]
        for _ in range(20):
            invocation = self.gi.workflows.show_invocation(workflow_id, invocation_id)
            if invocation["state"] == "scheduled":
                break

            time.sleep(0.5)

        invocation = self.gi.workflows.show_invocation(workflow_id, invocation_id)
        assert invocation["state"] == "scheduled"

    def test_invoke_workflow_parameters_normalized(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns_subworkflow.ga"))
        workflow_id = self.gi.workflows.import_workflow_from_local_path(path)["id"]
        history_id = self.gi.histories.create_history(name="TestWorkflowInvokeParametersNormalized")["id"]
        dataset_id = self._test_dataset(history_id)
        with pytest.raises(ConnectionError):
            self.gi.workflows.invoke_workflow(
                workflow_id, inputs={"0": {"src": "hda", "id": dataset_id}}, params={"1": {"1|2": "comma"}}
            )
        self.gi.workflows.invoke_workflow(
            workflow_id,
            inputs={"0": {"src": "hda", "id": dataset_id}},
            params={"1": {"1|2": "comma"}},
            parameters_normalized=True,
        )

    @test_util.skip_unless_galaxy("release_19.09")
    @test_util.skip_unless_tool("cat1")
    @test_util.skip_unless_tool("cat")
    def test_cancelling_workflow_scheduling(self):
        path = test_util.get_abspath(os.path.join("data", "test_workflow_pause.ga"))
        workflow = self.gi.workflows.import_workflow_from_local_path(path)
        workflow_id = workflow["id"]
        history_id = self.gi.histories.create_history(name="TestWorkflowState")["id"]
        dataset1_id = self._test_dataset(history_id)

        invocations = self.gi.workflows.get_invocations(workflow_id)
        assert len(invocations) == 0

        invocation = self.gi.workflows.invoke_workflow(
            workflow_id,
            inputs={"0": {"src": "hda", "id": dataset1_id}},
        )
        invocation_id = invocation["id"]
        invocations = self.gi.workflows.get_invocations(workflow_id)
        assert invocation_id in [inv["id"] for inv in invocations]

        invocation = self.gi.workflows.show_invocation(workflow_id, invocation_id)
        assert invocation["state"] in ["new", "ready"]

        invocation = self.gi.workflows.cancel_invocation(workflow_id, invocation_id)
        assert invocation["state"] in ["cancelled", "cancelling"]

    def test_import_export_workflow_from_local_path(self):
        with pytest.raises(TypeError):
            self.gi.workflows.import_workflow_from_local_path(None)  # type: ignore[arg-type]
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        imported_wf = self.gi.workflows.import_workflow_from_local_path(path)
        assert isinstance(imported_wf, dict)
        assert imported_wf["name"] == "paste_columns"
        assert imported_wf["url"].startswith("/api/workflows/")
        assert not imported_wf["deleted"]
        assert not imported_wf["published"]
        with pytest.raises(TypeError):
            self.gi.workflows.export_workflow_to_local_path(None, None, None)  # type: ignore[arg-type]
        export_dir = tempfile.mkdtemp(prefix="bioblend_test_")
        try:
            self.gi.workflows.export_workflow_to_local_path(imported_wf["id"], export_dir)
            dir_contents = os.listdir(export_dir)
            assert len(dir_contents) == 1
            export_path = os.path.join(export_dir, dir_contents[0])
            with open(export_path) as f:
                exported_wf_dict = json.load(f)
        finally:
            shutil.rmtree(export_dir)
        assert isinstance(exported_wf_dict, dict)

    def test_import_publish_workflow_from_local_path(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        imported_wf = self.gi.workflows.import_workflow_from_local_path(path, publish=True)
        assert isinstance(imported_wf, dict)
        assert not imported_wf["deleted"]
        assert imported_wf["published"]

    def test_import_export_workflow_dict(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        with open(path) as f:
            wf_dict = json.load(f)
        imported_wf = self.gi.workflows.import_workflow_dict(wf_dict)
        assert isinstance(imported_wf, dict)
        assert imported_wf["name"] == "paste_columns"
        assert imported_wf["url"].startswith("/api/workflows/")
        assert not imported_wf["deleted"]
        assert not imported_wf["published"]
        exported_wf_dict = self.gi.workflows.export_workflow_dict(imported_wf["id"])
        assert isinstance(exported_wf_dict, dict)

    def test_import_publish_workflow_dict(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        with open(path) as f:
            wf_dict = json.load(f)
        imported_wf = self.gi.workflows.import_workflow_dict(wf_dict, publish=True)
        assert isinstance(imported_wf, dict)
        assert not imported_wf["deleted"]
        assert imported_wf["published"]

    def test_get_workflows(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        workflow = self.gi.workflows.import_workflow_from_local_path(path)
        all_wfs = self.gi.workflows.get_workflows()
        assert len(all_wfs) > 0
        wfs_with_name = self.gi.workflows.get_workflows(name=workflow["name"])
        wf_list = [w for w in wfs_with_name if w["id"] == workflow["id"]]
        assert len(wf_list) == 1
        wf_data = wf_list[0]
        if "create_time" in workflow:  # Galaxy >= 20.01
            assert wf_data["create_time"] == workflow["create_time"]
        else:  # Galaxy <= 22.01
            assert wf_data["url"] == workflow["url"]

    def test_show_workflow(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        wf_data = self.gi.workflows.show_workflow(wf["id"])
        assert wf_data["id"] == wf["id"]
        assert wf_data["name"] == wf["name"]
        assert wf_data["url"] == wf["url"]
        assert len(wf_data["steps"]) == 3
        assert wf_data["inputs"] is not None

    def test_update_workflow_name(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        new_name = "new name"
        updated_wf = self.gi.workflows.update_workflow(wf["id"], name=new_name)
        assert updated_wf["name"] == new_name

    @test_util.skip_unless_galaxy("release_21.01")
    def test_update_workflow_published(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        assert not wf["published"]
        updated_wf = self.gi.workflows.update_workflow(wf["id"], published=True)
        assert updated_wf["published"]
        updated_wf = self.gi.workflows.update_workflow(wf["id"], published=False)
        assert not updated_wf["published"]

    @test_util.skip_unless_galaxy(
        "release_19.09"
    )  # due to Galaxy bug fixed in https://github.com/galaxyproject/galaxy/pull/9014
    def test_show_workflow_versions(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        wf_data = self.gi.workflows.show_workflow(wf["id"])
        assert wf_data["version"] == 0
        new_name = "new name"
        self.gi.workflows.update_workflow(wf["id"], name=new_name)
        updated_wf = self.gi.workflows.show_workflow(wf["id"])
        assert updated_wf["name"] == new_name
        assert updated_wf["version"] == 1
        updated_wf = self.gi.workflows.show_workflow(wf["id"], version=0)
        assert updated_wf["name"] == "paste_columns"
        assert updated_wf["version"] == 0
        updated_wf = self.gi.workflows.show_workflow(wf["id"], version=1)
        assert updated_wf["name"] == new_name
        assert updated_wf["version"] == 1

    @test_util.skip_unless_galaxy("release_19.09")
    def test_extract_workflow_from_history(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        history_id = self.gi.histories.create_history(name="test_wf_invocation")["id"]
        dataset1_id = self._test_dataset(history_id)
        dataset = {"src": "hda", "id": dataset1_id}
        invocation_id = self.gi.workflows.invoke_workflow(
            wf["id"],
            inputs={"Input 1": dataset, "Input 2": dataset},
            history_id=history_id,
            inputs_by="name",
        )["id"]
        invocation = self.gi.invocations.wait_for_invocation(invocation_id)
        wf1 = self.gi.workflows.show_workflow(invocation["workflow_id"])
        datasets = self.gi.histories.show_history(invocation["history_id"], contents=True)
        dataset_hids = [dataset["hid"] for dataset in datasets]
        job_ids = [step["job_id"] for step in invocation["steps"] if step["job_id"]]

        for job_id in job_ids:
            self.gi.jobs.wait_for_job(job_id)

        new_workflow_name = "My new workflow!"
        wf2 = self.gi.workflows.extract_workflow_from_history(
            history_id=invocation["history_id"],
            workflow_name=new_workflow_name,
            job_ids=job_ids,
            dataset_hids=dataset_hids,
        )
        wf2 = self.gi.workflows.show_workflow(wf2["id"])
        assert wf2["name"] == new_workflow_name
        assert len(wf1["steps"]) == len(wf2["steps"])
        for i in range(len(wf1["steps"])):
            assert wf1["steps"][str(i)]["type"] == wf2["steps"][str(i)]["type"]
            assert wf1["steps"][str(i)]["tool_id"] == wf2["steps"][str(i)]["tool_id"]

    def test_show_versions(self):
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        versions = self.gi.workflows.show_versions(wf["id"])
        assert len(versions) == 1
        version = versions[0]
        assert version["version"] == 0
        assert "update_time" in version
        assert "steps" in version

    @test_util.skip_unless_galaxy("release_21.01")
    def test_refactor_workflow(self):
        actions: List[Dict[str, Any]] = [
            {"action_type": "add_input", "type": "data", "label": "foo"},
            {"action_type": "update_step_label", "label": "bar", "step": {"label": "foo"}},
        ]
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        wf = self.gi.workflows.import_workflow_from_local_path(path)
        response = self.gi.workflows.refactor_workflow(wf["id"], actions, dry_run=True)
        assert len(response["action_executions"]) == len(actions)
        assert response["dry_run"] is True
        updated_steps = response["workflow"]["steps"]
        assert len(updated_steps) == 4
        assert {step["label"] for step in updated_steps.values()} == {"bar", None, "Input 1", "Input 2"}



"""
A base representation of an instance of Galaxy
"""

from typing import Optional

from bioblend.galaxy import (
    config,
    container_resolution,
    dataset_collections,
    datasets,
    datatypes,
    folders,
    forms,
    ftpfiles,
    genomes,
    groups,
    histories,
    invocations,
    jobs,
    libraries,
    quotas,
    roles,
    tool_data,
    tool_dependencies,
    tools,
    toolshed,
    users,
    visual,
    workflows,
)
from bioblend.galaxyclient import GalaxyClient


class GalaxyInstance(GalaxyClient):
    def __init__(
        self,
        url: str,
        key: Optional[str] = None,
        email: Optional[str] = None,
        password: Optional[str] = None,
        *,
        verify: bool = True,
    ) -> None:
        """
        A base representation of a connection to a Galaxy instance, identified
        by the server URL and user credentials.

        After you have created a ``GalaxyInstance`` object, access various
        modules via the class fields. For example, to work with histories and
        get a list of all the user's histories, the following should be done::

            from bioblend import galaxy

            gi = galaxy.GalaxyInstance(url='http://127.0.0.1:8000', key='your_api_key')

            hl = gi.histories.get_histories()

        :type url: str
        :param url: A FQDN or IP for a given instance of Galaxy. For example:
                    http://127.0.0.1:8080 . If a Galaxy instance is served under
                    a prefix (e.g., http://127.0.0.1:8080/galaxy/), supply the
                    entire URL including the prefix (note that the prefix must
                    end with a slash). If a Galaxy instance has HTTP Basic
                    authentication with username and password, then the
                    credentials should be included in the URL, e.g.
                    http://user:pass@host:port/galaxy/

        :type key: str
        :param key: User's API key for the given instance of Galaxy, obtained
                    from the user preferences. If a key is not supplied, an
                    email address and password must be and the key will
                    automatically be created for the user.

        :type email: str
        :param email: Galaxy e-mail address corresponding to the user.
                      Ignored if key is supplied directly.

        :type password: str
        :param password: Password of Galaxy account corresponding to the above
                         e-mail address. Ignored if key is supplied directly.

        :param verify: Whether to verify the server's TLS certificate
        :type verify: bool
        """
        super().__init__(url, key=key, email=email, password=password, verify=verify)
        self.libraries = libraries.LibraryClient(self)
        self.histories = histories.HistoryClient(self)
        self.workflows = workflows.WorkflowClient(self)
        self.invocations = invocations.InvocationClient(self)
        self.datasets = datasets.DatasetClient(self)
        self.dataset_collections = dataset_collections.DatasetCollectionClient(self)
        self.users = users.UserClient(self)
        self.genomes = genomes.GenomeClient(self)
        self.tools = tools.ToolClient(self)
        self.toolshed = toolshed.ToolShedClient(self)
        self.toolShed = self.toolshed  # historical alias
        self.config = config.ConfigClient(self)
        self.container_resolution = container_resolution.ContainerResolutionClient(self)
        self.visual = visual.VisualClient(self)
        self.quotas = quotas.QuotaClient(self)
        self.groups = groups.GroupsClient(self)
        self.roles = roles.RolesClient(self)
        self.datatypes = datatypes.DatatypesClient(self)
        self.jobs = jobs.JobsClient(self)
        self.forms = forms.FormsClient(self)
        self.ftpfiles = ftpfiles.FTPFilesClient(self)
        self.tool_data = tool_data.ToolDataClient(self)
        self.folders = folders.FoldersClient(self)
        self.tool_dependencies = tool_dependencies.ToolDependenciesClient(self)

    def __repr__(self) -> str:
        """
        A nicer representation of this GalaxyInstance object
        """
        return f"GalaxyInstance object for Galaxy at {self.base_url}"



import os
import sys
import tempfile

from common import get_one  # noqa:I100,I201

from bioblend.galaxy.objects import GalaxyInstance

# This is a "toy" example that should run much faster
# (once the cluster's resource manager allows it to run) than the
# real-world ones.  The workflow, which is imported from local disk,
# removes two header lines from a tabular file, then extracts one of
# the columns.  The input dataset is publicly available on CRS4's
# Orione Galaxy server.

URL = "https://orione.crs4.it"
API_KEY = os.getenv("GALAXY_API_KEY", "YOUR_API_KEY")
if API_KEY == "YOUR_API_KEY":
    sys.exit("API_KEY not set, see the README.txt file")
gi = GalaxyInstance(URL, api_key=API_KEY)

# import the workflow from the JSON dump

with open("small.ga") as f:
    wf = gi.workflows.import_new(f.read())

# Select the "Orione SupMat" library

library_name = "Orione SupMat"
library = get_one(gi.libraries.list(name=library_name))

# Select the input dataset

ds_name = "/RNA-Seq - Listeria monocytogenes/Listeria_monocytogenes_EGD_e_uid61583/NC_003210.rnt"
ld = get_one(library.get_datasets(name=ds_name))
input_map = {"input_tsv": ld}

# Run the workflow on a new history with the selected dataset as
# input, overriding the index of the column to remove; wait until the
# computation is complete.

history_name = "get_col output"
params = {"Cut1": {"columnList": "c2"}}
print(f"Running workflow: {wf.name} [{wf.id}]")
inv = wf.invoke(input_map, params=params, history=history_name, inputs_by="name")
out_hist = gi.histories.get(inv.history_id)
inv.wait()
print("Job has finished")
assert out_hist.name == history_name
print(f"Output history: {out_hist.name} [{out_hist.id}]")

# Save results to local disk
out_ds = get_one(out_hist.get_datasets(name="Cut on data 1"))
with tempfile.NamedTemporaryFile(prefix="bioblend_", delete=False) as tmp_f:
    out_ds.download(tmp_f)
print(f'Output downloaded to "{f.name}"')



"""
Contains possible interactions with the Galaxy Histories
"""

from typing import (
    Any,
    Dict,
    Literal,
    Optional,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class GenomeClient(Client):
    module = "genomes"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def get_genomes(self) -> list:
        """
        Returns a list of installed genomes

        :rtype: list
        :return: List of installed genomes
        """
        genomes = self._get()
        return genomes

    def show_genome(
        self,
        id: str,
        num: Optional[str] = None,
        chrom: Optional[str] = None,
        low: Optional[str] = None,
        high: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Returns information about build <id>

        :type id: str
        :param id: Genome build ID to use

        :type num: str
        :param num: num

        :type chrom: str
        :param chrom: chrom

        :type low: str
        :param low: low

        :type high: str
        :param high: high

        :rtype: dict
        :return: Information about the genome build
        """
        params: Dict[str, str] = {}
        if num:
            params["num"] = num
        if chrom:
            params["chrom"] = chrom
        if low:
            params["low"] = low
        if high:
            params["high"] = high
        return self._get(id=id, params=params)

    def install_genome(
        self,
        func: Literal["download", "index"] = "download",
        source: Optional[str] = None,
        dbkey: Optional[str] = None,
        ncbi_name: Optional[str] = None,
        ensembl_dbkey: Optional[str] = None,
        url_dbkey: Optional[str] = None,
        indexers: Optional[list] = None,
    ) -> Dict[str, Any]:
        """
        Download and/or index a genome.

        :type func: str
        :param func: Allowed values: 'download', Download and index; 'index', Index only

        :type source: str
        :param source: Data source for this build. Can be: UCSC, Ensembl, NCBI, URL

        :type dbkey: str
        :param dbkey: DB key of the build to download, ignored unless 'UCSC' is specified as the source

        :type ncbi_name: str
        :param ncbi_name: NCBI's genome identifier, ignored unless NCBI is specified as the source

        :type ensembl_dbkey: str
        :param ensembl_dbkey: Ensembl's genome identifier, ignored unless Ensembl is specified as the source

        :type url_dbkey: str
        :param url_dbkey: DB key to use for this build, ignored unless URL is specified as the source

        :type indexers: list
        :param indexers: POST array of indexers to run after downloading (indexers[] = first, indexers[] = second, ...)

        :rtype: dict
        :return: dict( status: 'ok', job: <job ID> )
                 If error:
                 dict( status: 'error', error: <error message> )
        """
        payload: Dict[str, Any] = {}
        if source:
            payload["source"] = source
        if func:
            payload["func"] = func
        if dbkey:
            payload["dbkey"] = dbkey
        if ncbi_name:
            payload["ncbi_name"] = ncbi_name
        if ensembl_dbkey:
            payload["ensembl_dbkey"] = ensembl_dbkey
        if url_dbkey:
            payload["url_dbkey"] = url_dbkey
        if indexers:
            payload["indexers"] = indexers
        return self._post(payload)



"""
Contains interactions dealing with Galaxy dependency resolvers.
"""

from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
    TYPE_CHECKING,
)

from bioblend.galaxy.client import Client

if TYPE_CHECKING:
    from bioblend.galaxy import GalaxyInstance


class ToolDependenciesClient(Client):
    module = "dependency_resolvers"

    def __init__(self, galaxy_instance: "GalaxyInstance") -> None:
        super().__init__(galaxy_instance)

    def summarize_toolbox(
        self,
        index: Optional[int] = None,
        tool_ids: Optional[List[str]] = None,
        resolver_type: Optional[str] = None,
        include_containers: bool = False,
        container_type: Optional[str] = None,
        index_by: Literal["requirements", "tools"] = "requirements",
    ) -> list:
        """
        Summarize requirements across toolbox (for Tool Management grid).

        :type index: int
        :param index: index of the dependency resolver with respect to
            the dependency resolvers config file

        :type tool_ids: list
        :param tool_ids: tool_ids to return when index_by=tools

        :type resolver_type: str
        :param resolver_type: restrict to specified resolver type

        :type include_containers: bool
        :param include_containers: include container resolvers in resolution

        :type container_type: str
        :param container_type: restrict to specified container type

        :type index_by: str
        :param index_by: By default results are grouped by requirements.  Set to 'tools'
          to return one entry per tool.

        :rtype: list of dicts
        :returns: dictified descriptions of the dependencies, with attribute
          `dependency_type: None` if no match was found.
          For example::

            [{'requirements': [{'name': 'galaxy_sequence_utils',
                                'specs': [],
                                'type': 'package',
                                'version': '1.1.4'},
                               {'name': 'bx-python',
                                'specs': [],
                                'type': 'package',
                                'version': '0.8.6'}],
              'status': [{'cacheable': False,
                          'dependency_type': None,
                          'exact': True,
                          'model_class': 'NullDependency',
                          'name': 'galaxy_sequence_utils',
                          'version': '1.1.4'},
                          {'cacheable': False,
                          'dependency_type': None,
                          'exact': True,
                          'model_class': 'NullDependency',
                          'name': 'bx-python',
                          'version': '0.8.6'}],
              'tool_ids': ['vcf_to_maf_customtrack1']}]

        .. note::
          This method works only on Galaxy 20.01 or later and if the user is a
          Galaxy admin. It relies on an experimental API particularly tied to
          the GUI and therefore is subject to breaking changes.
        """
        assert index_by in ["tools", "requirements"], "index_by must be one of 'tools' or 'requirements'."
        params = {
            "include_containers": str(include_containers),
            "index_by": index_by,
        }
        if index:
            params["index"] = str(index)
        if tool_ids:
            params["tool_ids"] = ",".join(tool_ids)
        if resolver_type:
            params["resolver_type"] = resolver_type
        if container_type:
            params["container_type"] = container_type

        url = "/".join((self._make_url(), "toolbox"))
        return self._get(url=url, params=params)

    def unused_dependency_paths(self) -> List[str]:
        """
        List unused dependencies
        """
        url = "/".join((self._make_url(), "unused_paths"))
        return self._get(url=url)

    def delete_unused_dependency_paths(self, paths: List[str]) -> None:
        """
        Delete unused paths

        :type paths: list
        :param paths: paths to delete

        """
        payload: Dict[str, Any] = {"paths": paths}
        url = "/".join((self._make_url(), "unused_paths"))
        self._put(url=url, payload=payload)



import json
import os
import sys

from common import get_one  # noqa:I100,I201

from bioblend.galaxy.objects import GalaxyInstance

URL = "https://orione.crs4.it"
API_KEY = os.getenv("GALAXY_API_KEY", "YOUR_API_KEY")
if API_KEY == "YOUR_API_KEY":
    sys.exit("API_KEY not set, see the README.txt file")
gi = GalaxyInstance(URL, api_key=API_KEY)

# Select "W5 - Metagenomics" from published workflows

workflow_name = "W5 - Metagenomics"
previews = gi.workflows.get_previews(name=workflow_name, published=True)
p = get_one(_ for _ in previews if _.published)

# Import the workflow to user space

iw = gi.workflows.import_shared(p.id)

# Create a new history

history_name = f"{workflow_name} output"
h = gi.histories.create(history_name)

# Select the "Orione SupMat" library

library_name = "Orione SupMat"
library = get_one(gi.libraries.list(name=library_name))

# Select the "/Metagenomics/MetagenomicsDataset.fq" dataset

ds_name = "/Metagenomics/MetagenomicsDataset.fq"
input_map = {"Input Dataset": h.import_dataset(get_one(library.get_datasets(name=ds_name)))}

# Select the blastn step

tool_id = "toolshed.g2.bx.psu.edu/repos/devteam/ncbi_blast_plus/ncbi_blastn_wrapper/0.1.00"
step_id = get_one(iw.tool_labels_to_ids[tool_id])
ws = iw.steps[step_id]

# Get (a copy of) the parameters dict for the selected step

ws_parameters = ws.tool_inputs.copy()

# Run the workflow on a new history with the selected dataset
# as input, setting the BLAST db to "16SMicrobial-20131106"

params = {tool_id: {"db_opts": json.loads(ws_parameters["db_opts"])}}
params[tool_id]["db_opts"]["database"] = "16SMicrobial-20131106"
inv = iw.invoke(input_map, params=params, history=h, inputs_by="name")
out_hist = gi.histories.get(inv.history_id)
assert out_hist.name == history_name

print(f"Running workflow: {iw.name} [{iw.id}]")
print(f"Output history: {out_hist.name} [{out_hist.id}]")



import contextlib
import os
import time
from typing import (
    Any,
    Dict,
)

from . import (
    GalaxyTestBase,
    test_util,
)


class TestGalaxyInvocations(GalaxyTestBase.GalaxyTestBase):
    def setUp(self):
        super().setUp()
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        self.workflow_id = self.gi.workflows.import_workflow_from_local_path(path)["id"]
        self.history_id = self.gi.histories.create_history(name="TestGalaxyInvocations")["id"]
        self.dataset_id = self._test_dataset(self.history_id)
        path = test_util.get_abspath(os.path.join("data", "test_workflow_pause.ga"))
        self.pause_workflow_id = self.gi.workflows.import_workflow_from_local_path(path)["id"]

    def tearDown(self):
        self.gi.histories.delete_history(self.history_id, purge=True)

    @test_util.skip_unless_galaxy("release_19.09")
    def test_cancel_invocation(self):
        invocation = self._invoke_pause_workflow()
        invocation_id = invocation["id"]
        invocations = self.gi.invocations.get_invocations()
        assert invocation_id in [inv["id"] for inv in invocations]

        invocation = self.gi.invocations.show_invocation(invocation_id)
        assert invocation["state"] in ["new", "ready"]

        invocation = self.gi.invocations.cancel_invocation(invocation_id)
        assert invocation["state"] in ["cancelled", "cancelling"]

    @test_util.skip_unless_galaxy("release_20.01")
    def test_get_invocations(self):
        invoc1 = self._invoke_workflow()

        # Run the first workflow on another history
        dataset = {"src": "hda", "id": self.dataset_id}
        hist2_id = self.gi.histories.create_history("hist2")["id"]
        invoc2 = self.gi.workflows.invoke_workflow(
            self.workflow_id, history_id=hist2_id, inputs={"Input 1": dataset, "Input 2": dataset}, inputs_by="name"
        )

        # Run another workflow on the 2nd history
        path = test_util.get_abspath(os.path.join("data", "paste_columns.ga"))
        workflow2_id = self.gi.workflows.import_workflow_from_local_path(path)["id"]
        invoc3 = self.gi.workflows.invoke_workflow(
            workflow2_id, history_id=hist2_id, inputs={"Input 1": dataset, "Input 2": dataset}, inputs_by="name"
        )

        for invoc in (invoc1, invoc2, invoc3):
            self.gi.invocations.wait_for_invocation(invoc["id"])

        # Test filtering by workflow ID
        for wf_id, expected_invoc_num in {self.workflow_id: 2, workflow2_id: 1}.items():
            invocs = self.gi.invocations.get_invocations(workflow_id=wf_id)
            assert len(invocs) == expected_invoc_num
            for invoc in invocs:
                assert invoc["workflow_id"] == wf_id

        # Test filtering by history ID
        for hist_id, expected_invoc_num in {self.history_id: 1, hist2_id: 2}.items():
            invocs = self.gi.invocations.get_invocations(history_id=hist_id)
            assert len(invocs) == expected_invoc_num
            for invoc in invocs:
                assert invoc["history_id"] == hist_id

        # Test limiting
        limit_invocs = self.gi.invocations.get_invocations(limit=2)
        assert len(limit_invocs) == 2

        self.gi.histories.delete_history(hist2_id, purge=True)

    @test_util.skip_unless_galaxy("release_19.09")
    def test_get_invocation_report(self):
        invocation = self._invoke_workflow()

        invocation_id = invocation["id"]
        workflow_id = invocation["workflow_id"]
        report = self.gi.invocations.get_invocation_report(invocation_id)
        assert report["workflows"] == {workflow_id: {"name": "paste_columns"}}
        with contextlib.suppress(Exception):
            # This can fail if dependencies as weasyprint are not installed on the Galaxy server
            ret = self.gi.invocations.get_invocation_report_pdf(invocation_id, "report.pdf")
            assert ret is None

    @test_util.skip_unless_galaxy("release_20.09")
    def test_get_invocation_biocompute_object(self):
        invocation = self._invoke_workflow()

        self.gi.invocations.wait_for_invocation(invocation["id"])
        biocompute_object = self.gi.invocations.get_invocation_biocompute_object(invocation["id"])
        assert len(biocompute_object["description_domain"]["pipeline_steps"]) == 1

    @test_util.skip_unless_galaxy("release_19.09")
    def test_get_invocation_jobs_summary(self):
        invocation = self._invoke_workflow()
        self.gi.invocations.wait_for_invocation(invocation["id"])
        jobs_summary = self.gi.invocations.get_invocation_summary(invocation["id"])
        assert jobs_summary["populated_state"] == "ok"
        step_jobs_summary = self.gi.invocations.get_invocation_step_jobs_summary(invocation["id"])
        assert len(step_jobs_summary) == 1
        assert step_jobs_summary[0]["populated_state"] == "ok"

    @test_util.skip_unless_galaxy("release_19.09")
    @test_util.skip_unless_tool("cat1")
    @test_util.skip_unless_tool("cat")
    def test_workflow_scheduling(self):
        invocation = self._invoke_pause_workflow()
        invocation_id = invocation["id"]

        def invocation_steps_by_order_index() -> Dict[int, Dict[str, Any]]:
            invocation = self.gi.invocations.show_invocation(invocation_id)
            return {s["order_index"]: s for s in invocation["steps"]}

        for _ in range(20):
            if 2 in invocation_steps_by_order_index():
                break
            time.sleep(0.5)
        steps = invocation_steps_by_order_index()
        pause_step = steps[2]
        assert self.gi.invocations.show_invocation_step(invocation_id, pause_step["id"])["action"] is None
        self.gi.invocations.run_invocation_step_action(invocation_id, pause_step["id"], action=True)
        assert self.gi.invocations.show_invocation_step(invocation_id, pause_step["id"])["action"]
        self.gi.invocations.wait_for_invocation(invocation["id"])

    @test_util.skip_unless_galaxy("release_21.01")
    def test_rerun_invocation(self):
        invocation = self._invoke_workflow()
        self.gi.invocations.wait_for_invocation(invocation["id"])
        rerun_invocation = self.gi.invocations.rerun_invocation(invocation["id"], import_inputs_to_history=True)
        self.gi.invocations.wait_for_invocation(rerun_invocation["id"])
        history = self.gi.histories.show_history(rerun_invocation["history_id"], contents=True)
        assert len(history) == 3

    def _invoke_workflow(self) -> Dict[str, Any]:
        dataset = {"src": "hda", "id": self.dataset_id}

        return self.gi.workflows.invoke_workflow(
            self.workflow_id,
            inputs={"Input 1": dataset, "Input 2": dataset},
            history_id=self.history_id,
            inputs_by="name",
        )

    def _invoke_pause_workflow(self) -> Dict[str, Any]:
        return self.gi.workflows.invoke_workflow(
            self.pause_workflow_id,
            inputs={"0": {"src": "hda", "id": self.dataset_id}},
            history_id=self.history_id,
        )
